"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import re
import sys
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Generic

from typing_extensions import TypeVar

from ._const import URL, impl_eq_via_id
from .app import App, PartialApp, PartialAppPriceOverview
from .enums import Currency, Language, LicenseFlag, LicenseType, PaymentMethod
from .errors import HTTPException
from .models import CDNAsset
from .types.id import DepotID, Intable, PackageID
from .utils import DateTime

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Mapping

    from .manifest import AppInfo, Depot, HeadlessDepot, Manifest, PackageInfo
    from .protobufs.client_server import CMsgClientLicenseListLicense
    from .state import ConnectionState
    from .store import PackageStoreItem
    from .types import app, package
    from .user import ClientUser, User


__all__ = (
    "Package",
    "FetchedPackage",
    "FetchedAppPackage",
    "License",
)

NameT = TypeVar("NameT", bound=str | None, default=str | None, covariant=True)
AppT = TypeVar("AppT", bound=App, covariant=True)


@impl_eq_via_id
class Package(Generic[NameT]):
    """Represents a package, a collection of one or more apps and depots.

    Read more on `steamworks <https://partner.steamgames.com/doc/store/application/packages>`_.
    """

    __slots__ = (
        "id",
        "name",
    )

    def __init__(self, *, id: Intable, name: NameT = None):
        self.id = PackageID(int(id))
        """The package's ID."""
        self.name = name
        """The package's name."""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, name={self.name!r})"

    @property
    def url(self) -> str:
        """The package's store page URL."""
        return f"{URL.STORE}/sub/{self.id}"


class PartialPackage(Package[NameT]):
    """A package with state."""

    __slots__ = ("_state",)

    def __init__(self, state: ConnectionState, id: Intable, name: NameT = None):
        super().__init__(id=id, name=name)
        self._state = state

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id}>"

    async def apps(self, *, language: Language | None = None) -> list[PartialApp[str]]:
        """Fetches this package's apps."""
        fetched = await self.fetch(language=language)
        return fetched._apps

    async def fetch(self, *, language: Language | None = None) -> FetchedPackage:
        """Fetches this package's information.

        Shorthand for:

        .. code:: python

            package = await client.fetch_package(package, language=language)
        """
        return await self._state.client.fetch_package(self.id, language=language)

    async def info(self) -> PackageInfo:
        """Fetches this package's product info.

        Shorthand for:

        .. code:: python

            (info,) = await client.fetch_product_info(packages=[package])
        """
        _, (info,) = await self._state.fetch_product_info(package_ids=(self.id,))
        return info

    async def store_item(self, *, language: Language | None = None) -> PackageStoreItem:
        """Fetches this package's store item.

        Shorthand for:

        .. code:: python

            (item,) = await client.fetch_store_item(packages=[package], language=language)
        """
        from .store import PackageStoreItem

        language = language or self._state.http.language
        (item,) = await self._state.fetch_store_info(package_ids=(self.id,), language=language)
        return PackageStoreItem(self._state, item, language)

    async def apps_info(self) -> list[AppInfo]:
        """Fetch the product info for all apps in this package.

        Shorthand for:

        .. code:: python

            infos = await client.fetch_product_info(apps=await package.apps())
        """
        apps = await self.apps()
        if not apps:
            return []
        infos, _ = await self._state.fetch_product_info(app.id for app in apps)
        return infos

    async def depots(self) -> list[Depot | HeadlessDepot]:
        """Fetches this package's depots."""
        try:
            depot_ids: list[DepotID] = self.depot_ids  # type: ignore
        except AttributeError:
            info = await self.info()
            depot_ids = info.depot_ids

        apps_info = await self.apps_info()

        return [
            depot
            for app_info in apps_info
            for branch in app_info._branches.values()
            for depot in branch.depots
            if depot.id in depot_ids
        ] + [depot for app_info in apps_info for depot in app_info.headless_depots if depot.id in depot_ids]

    async def manifests(
        self,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
        branch: str = "public",
        passwords: Mapping[
            AppT, str | None
        ] = {},  # workaround for no covariant in the key mapping. Shouldn't fall victim to https://github.com/python/typing/pull/273
        password_hashes: Mapping[AppT, str] = {},
    ) -> AsyncGenerator[Manifest, None]:
        """An :term:`asynchronous iterator` for accessing this package's :class:`steam.Manifest`\\s.

        Examples
        --------

        Usage:

        .. code:: python

            async for manifest in package.manifests(limit=10):
                print("Manifest:", manifest.name)
                print(f"Contains {len(manifest.paths)} manifests")

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of :class:`.Manifests` to return.
        before
            The time to get manifests before.
        after
            The time to get manifests after.
        branch
            The name of the branch to fetch manifests from.
        passwords
            A mapping of the password for a branch by its app, if any.
        password_hashes
            A mapping of the hashed password for a branch by its app, if any.

        Yields
        ------
        :class:`Manifest`
        """
        try:
            depot_ids: list[DepotID] = self.depot_ids  # type: ignore
        except AttributeError:
            info = await self.info()
            depot_ids = info.depot_ids

        for app in await self.apps():
            async for manifest in app.manifests(
                limit=limit,
                before=before,
                after=after,
                branch=branch,
                password=passwords.get(app),  # type: ignore
                password_hash=password_hashes.get(app, ""),  # type: ignore
            ):
                if manifest.depot_id not in depot_ids:
                    continue

                yield manifest

                if limit is not None:
                    limit -= 1
                if limit == 0:
                    return

    async def fetch_manifest(
        self, *, id: int, depot_id: int, branch: str = "public", password_hash: str = ""
    ) -> Manifest:
        """Fetch a manifest from one of the :meth:`apps` manifests.

        Parameters
        ----------
        id
            The ID of the manifest to fetch.
        depot_id
            The ID of the depot the manifest is from.
        branch
            The name of the branch the manifest is from.
        password_hash
            The hashed password for the manifest.
        """
        err = RuntimeError("There are no apps in this package")
        if sys.version_info >= (3, 11):
            errs: list[HTTPException] = []
        for app in await self.apps():
            try:
                return await app.fetch_manifest(id=id, depot_id=depot_id, branch=branch, password_hash=password_hash)
            except HTTPException as e:
                if sys.version_info >= (3, 11):
                    errs.append(e)
                else:
                    err = e
        if sys.version_info >= (3, 11):
            raise ExceptionGroup("No manifest found with id and depot_id supplied", errs)
        else:
            raise err

    async def redeem(self) -> License:
        """Redeem this package if it's a free-on-demand package."""
        return await self._state.client.redeem_package(self.id)


@dataclass(slots=True)
class PackagePriceOverview(PartialAppPriceOverview):
    individual: int


class FetchedPackage(PartialPackage[str]):
    """Represents a package that was fetched from steam."""

    def __init__(self, state: ConnectionState, data: package.FetchedPackage):
        super().__init__(state, name=data["name"], id=data["packageid"])
        self._apps = [PartialApp(state, id=app["id"], name=app["name"]) for app in data["apps"]]
        self.description = data["page_content"]
        self.created_at = DateTime.parse_steam_date(data["release_date"]["date"], full_month=False)
        self.logo = CDNAsset(state, data["header_image"])
        platforms = data["platforms"]
        self._on_windows = platforms["windows"]
        self._on_mac_os = platforms["mac"]
        self._on_linux = platforms["linux"]
        currency = Currency[data["price"].pop("currency")]  # type: ignore
        self.price_overview = PackagePriceOverview(currency=currency, **data["price"])  # type: ignore

    def is_free(self) -> bool:
        """Whether the package is free."""
        return self.price_overview.final == 0

    def is_on_windows(self) -> bool:
        """Whether the package is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the package is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the package is playable on Linux."""
        return self._on_linux


@dataclass(slots=True)
class FetchedAppPackagePriceOverview:
    percent_discount: int
    final: int


class FetchedAppPackage(PartialPackage[str]):
    __slots__ = (
        "_is_free",
        "price_overview",
    )

    def __init__(self, state: ConnectionState, data: app.PackageGroupSub):
        name, _, _ = data["option_text"].rpartition(" - ")
        super().__init__(state, name=name, id=data["packageid"])
        self._is_free = data["is_free_license"]
        self.price_overview = FetchedAppPackagePriceOverview(
            (
                int(
                    re.search(r"-?(\d)", savings)[0],  # type: ignore
                )
                if (savings := data["percent_savings_text"].strip())
                else 0
            ),
            data["price_in_cents_with_discount"],  # this isn't always in cents
        )

    def is_free(self) -> bool:
        return self._is_free


class License(PartialPackage[None]):
    """Represents a License to a package the client user has access to."""

    __slots__ = (
        "owner",
        "type",
        "flags",
        "created_at",
        "master_package",
        "next_process_at",
        "time_limit",
        "time_used",
        "payment_method",
        "purchase_country_code",
        "territory_code",
        "change_number",
        "initial_period",
        "initial_time_unit",
        "renewal_period",
        "renewal_time_unit",
        "access_token",
    )

    def __init__(self, state: ConnectionState, proto: CMsgClientLicenseListLicense, owner: ClientUser | User):
        super().__init__(state, id=proto.package_id)
        self.owner = owner
        """The license's owner."""
        self.type = LicenseType.try_value(proto.license_type)
        """The license's type."""
        self.flags = LicenseFlag.try_value(proto.flags)
        """The license's flags."""
        self.created_at = DateTime.from_timestamp(proto.time_created) if proto.time_created else None
        """The license's creation date."""
        self.master_package = PartialPackage(state, id=proto.master_package_id) if proto.master_package_id else None
        """The license's master package."""
        self.next_process_at = DateTime.from_timestamp(proto.time_next_process)
        """The date when the license will be processed."""
        self.time_limit = timedelta(minutes=proto.minute_limit) if proto.minute_limit else None
        """The time limit for the license."""
        self.time_used = timedelta(minutes=proto.minutes_used)
        """The time the license has been used."""
        self.payment_method = PaymentMethod.try_value(proto.payment_method)
        """The payment method used for the license."""
        self.purchase_country_code = proto.purchase_country_code
        """The country code of the license's purchase."""
        self.territory_code = proto.territory_code
        """The license's territory code."""
        self.change_number = proto.change_number
        """The license's current change number."""
        self.initial_period = proto.initial_period
        """The license's initial period."""
        self.initial_time_unit = proto.initial_time_unit
        """The license's initial time unit."""
        self.renewal_period = timedelta(minutes=proto.renewal_period)
        """The license's renewal period."""
        self.renewal_time_unit = proto.renewal_time_unit
        """The license's renewal time unit."""
        self.access_token = proto.access_token
        """The license's access token."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} owner={self.owner!r} flags={self.flags!r}>"

    @property
    def time_remaining(self) -> timedelta | None:
        """The amount of time that this license can be used for."""
        if self.flags & LicenseFlag.Expired > 0:
            return
        if self.time_limit is None:
            return
        return self.time_limit - self.time_used

    async def remove(self) -> None:
        """Remove the license from the client user's account."""
        await self._state.http.remove_license(self.id)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar, Generic, TypeAlias

from typing_extensions import TypeVar

from ._const import _HasChatGroupMixin

if TYPE_CHECKING:
    from .abc import PartialUser
    from .app import PartialApp
    from .clan import Clan
    from .enums import FriendRelationship
    from .game_server import GameServer
    from .group import Group
    from .state import ConnectionState
    from .user import ClientUser, User


__all__ = (
    "Invite",
    "UserInvite",
    "ClanInvite",
    "GroupInvite",
    "AppInvite",
)


@dataclass(slots=True)
class _Invite(metaclass=ABCMeta):
    _state: ConnectionState
    user: User | ClientUser | PartialUser
    """The user that was invited."""
    author: User | ClientUser | PartialUser
    """The user who sent the invite."""
    relationship: FriendRelationship | None
    """The relationship you have with the invitee."""
    REPR_ATTRS: ClassVar[tuple[str, ...]]

    @abstractmethod
    async def accept(self) -> None:
        """Accept the invite request."""
        raise NotImplementedError()

    def __repr__(self) -> str:
        cls = self.__class__
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in cls.REPR_ATTRS]
        return f"<{cls.__name__} {' '.join(resolved)}>"


@dataclass(repr=False, slots=True)
class UserInvite(_Invite):
    """Represents a invite from a user to become their friend."""

    REPR_ATTRS: ClassVar = ("author", "relationship")

    async def accept(self) -> None:
        await self._state.add_user(self.author.id64)

    async def decline(self) -> None:
        """Decline the invite request."""
        await self._state.remove_user(self.author.id64)


ClanT = TypeVar("ClanT", bound="Clan | None", default="Clan | None", covariant=True)
GroupT = TypeVar("GroupT", bound="Group | None", default="Group | None", covariant=True)


@dataclass(repr=False, slots=True)
class _ChatGroupInvite(_Invite, _HasChatGroupMixin, Generic[ClanT, GroupT]):
    if TYPE_CHECKING:
        code: str | None
        clan: ClanT
        group: GroupT

    async def revoke(self) -> None:
        """Revoke the invite request."""
        await self._state.revoke_chat_group_invite(self.user.id64, self._chat_group._id)


@dataclass(repr=False, slots=True)
class ClanInvite(_ChatGroupInvite["Clan", None]):
    """Represents an invitation to join a :class:`~steam.Clan` from a user."""

    REPR_ATTRS: ClassVar = ("author", "relationship", "clan")
    clan: Clan
    """The clan to join."""
    code: str | None = None
    group: None = None

    async def accept(self) -> None:
        await self._state.respond_to_clan_invite(self.clan.id64, True)
        await self.clan.join()

    async def decline(self) -> None:
        """Decline the invite request."""
        await self._state.respond_to_clan_invite(self.clan.id64, False)


@dataclass(repr=False, slots=True)
class GroupInvite(_ChatGroupInvite[None, "Group"]):
    """Represents an invitation from a user to join a group."""

    REPR_ATTRS: ClassVar = ("author", "relationship", "group")
    group: Group
    """The group to join."""
    code: str | None = None
    clan: None = None

    async def accept(self) -> None:
        await self.group.join(invite_code=self.code)


ChatGroupInvite: TypeAlias = ClanInvite | GroupInvite


@dataclass(repr=False, slots=True)
class AppInvite(_Invite):
    """Represents an invitation to join a :class:`~steam.App` from a user."""

    REPR_ATTRS: ClassVar = ("author", "app")
    app: PartialApp
    """The app to join."""
    server: GameServer | None
    """The server you being requested to join"""
    connect: str
    """A string containing information about the invite."""

    async def accept(self) -> None:
        return await super().accept()


Invite: TypeAlias = UserInvite | ChatGroupInvite | AppInvite
"""Represents an invite from a user to become their friend, join a clan, join a group, or join an app."""



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING, cast, overload

from ._const import _HasChatGroupMixin
from .enums import Type
from .protobufs import chat
from .types.id import RoleID

if TYPE_CHECKING:
    from typing_extensions import Self

    from .chat import ChatGroup, Member
    from .clan import Clan
    from .group import Group
    from .state import ConnectionState

__all__ = (
    "Role",
    "RolePermissions",
)


class Role(_HasChatGroupMixin):
    """Represents a role in a chat group."""

    __slots__ = ("id", "name", "ordinal", "clan", "group", "permissions", "_state")
    group: Group | None
    """The group the role belongs to, if any."""
    clan: Clan | None
    """The clan the role belongs to, if any."""

    def __init__(self, state: ConnectionState, group: ChatGroup, role: chat.Role, permissions: chat.RoleActions):
        self._state = state
        self.id = RoleID(role.role_id)
        """The ID of the role."""
        self.name = role.name.removeprefix("#ChatRoomRole_")
        """The name of the role."""
        self.ordinal = role.ordinal
        """The ordinal of the role."""

        if group.type is Type.Clan:  # these are reasonably safe casts as ChatGroup is an abc
            self.clan = cast("Clan", group)
            self.group = None
        else:
            self.group = cast("Group", group)
            self.clan = None
        self.permissions = RolePermissions(permissions)
        """The permissions of the role."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id}>"

    @property
    def members(self) -> list[Member]:
        """The members that have this role."""
        return [m for m in self._chat_group.members if self in m.roles]

    async def edit(
        self, *, name: str | None = None, permissions: RolePermissions | None = None, ordinal: int | None = None
    ) -> None:
        """Edit this role.

        Parameters
        ----------
        name
            The new name of the role.
        permissions
            The new permissions of the role.
        ordinal
            The new ordinal (position) of the role.
        """
        await self._state.edit_role(self._chat_group._id, self.id, name=name, permissions=permissions, ordinal=ordinal)

    async def delete(self) -> None:
        """Delete this role."""
        await self._state.delete_role(self._chat_group._id, self.id)


class RolePermissions:
    """Represents the permissions of a role in a chat group."""

    __slots__ = (
        "kick",
        "ban_members",
        "invite",
        "manage_group",
        "send_messages",
        "read_message_history",
        "change_group_roles",
        "change_user_roles",
        "mention_all",
        "set_watching_broadcast",
    )

    def __init__(self, proto: chat.RoleActions):
        self.kick = proto.can_kick
        """Whether the role can kick members."""
        self.ban_members = proto.can_ban
        """Whether the role can ban members."""
        self.invite = proto.can_invite
        """Whether the role can invite members."""
        self.manage_group = proto.can_change_tagline_avatar_name
        """Whether the role can manage the group."""
        self.send_messages = proto.can_chat
        """Whether the role can send messages."""
        self.read_message_history = proto.can_view_history
        """Whether the role can read message history."""
        self.change_group_roles = proto.can_change_group_roles
        """Whether the role can change group roles."""
        self.change_user_roles = proto.can_change_user_roles
        """Whether the role can change user roles."""
        self.mention_all = proto.can_mention_all
        """Whether the role can mention all."""
        self.set_watching_broadcast = proto.can_set_watching_broadcast
        """Whether the role can set watching broadcast."""

    def copy(self) -> Self:
        return self.__class__(self.to_proto())

    __copy__ = copy

    def to_proto(self) -> chat.RoleActions:
        return chat.RoleActions(
            can_kick=self.kick,
            can_ban=self.ban_members,
            can_invite=self.invite,
            can_change_tagline_avatar_name=self.manage_group,
            can_chat=self.send_messages,
            can_view_history=self.read_message_history,
            can_change_group_roles=self.change_group_roles,
            can_change_user_roles=self.change_user_roles,
            can_mention_all=self.mention_all,
            can_set_watching_broadcast=self.set_watching_broadcast,
        )

    @overload
    def replace(  # type: ignore
        self,
        *,
        kick: bool = ...,
        ban_members: bool = ...,
        invite: bool = ...,
        manage_group: bool = ...,
        send_messages: bool = ...,
        read_message_history: bool = ...,
        change_group_roles: bool = ...,
        change_user_roles: bool = ...,
        mention_all: bool = ...,
        set_watching_broadcast: bool = ...,
    ) -> Self: ...

    def replace(self, **kwargs: bool) -> Self:
        """Return a new RolePermissions with the specified changes."""
        new = self.copy()
        for name, value in kwargs.items():
            setattr(new, name, value)
        return new



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import itertools
import re
from datetime import date, datetime, timezone
from ipaddress import IPv4Address
from typing import TYPE_CHECKING, Literal, TypeVar, cast, overload

from bs4 import BeautifulSoup, Tag
from typing_extensions import Self
from yarl import URL

from . import utils
from ._const import HTML_PARSER, UNIX_EPOCH
from .abc import Commentable, PartialUser, _CommentableKwargs, _CommentThreadType
from .app import App, PartialApp
from .channel import ClanChannel
from .chat import ChatGroup, Member, PartialMember
from .enums import ClanAccountFlags, EventType, Language, Type, UserNewsType
from .event import Announcement, Event
from .id import ID, parse_id64
from .protobufs import chat
from .types.id import ID32, ID64, Intable
from .utils import BBCodeStr, DateTime, parse_bb_code

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Iterable, Sequence

    from .media import Media
    from .state import ConnectionState
    from .types.http import IPAdress
    from .user import User

__all__ = (
    "Clan",
    "ClanMember",
)

BoringEvents = Literal[
    EventType.Other,
    EventType.Chat,
    EventType.Party,
    EventType.Meeting,
    EventType.SpecialCause,
    EventType.MusicAndArts,
    EventType.Sports,
    EventType.Trip,
    # EventType.Broadcast,  # TODO need to wait until implementing stream support for this
]
CreateableEvents = BoringEvents | Literal[EventType.Game]

BoringEventT = TypeVar("BoringEventT", bound=BoringEvents)


class ClanMember(Member["Clan", None]):
    def __init__(self, state: ConnectionState, clan: Clan, user: User, proto: chat.Member):
        super().__init__(state, clan, user, proto)
        self.clan = clan


class PartialClan(ID[Literal[Type.Clan]], Commentable):
    def __init__(self, state: ConnectionState, id: Intable):
        super().__init__(id, type=Type.Clan)
        self._state = state

    @property
    def _commentable_kwargs(self) -> _CommentableKwargs:
        return {
            "id64": self.id64,
        }

    @utils.classproperty
    def _COMMENTABLE_TYPE(cls: type[Self]) -> _CommentThreadType:  # type: ignore
        return _CommentThreadType.Clan

    async def edit(
        self,
        *,
        abbreviation: str | None = None,
        headline: str | None = None,
        summary: str | None = None,
        community_url: URL | str | None = None,
        language: Language | None = None,
        country: str | None = None,
        state: str | None = None,
        city: str | None = None,
        apps: Iterable[App] | None = None,
        avatar: Media | None = None,
    ) -> None:
        """Edits the clan.

        Parameters
        ----------
        abbreviation
            The new abbreviation for the clan.
        headline
            The new headline for the clan.
        summary
            The new summary for the clan.
        community_url
            The new custom URL for the clan.
        language
            The new language for the clan.
        country
            The new country for the clan.
        state
            The new state for the clan.
        city
            The new city for the clan.
        apps
            The new apps for the clan.
        avatar
            The new avatar for the clan.
        """
        if isinstance(community_url, str):
            community_url = URL(community_url)
        if community_url is not None:
            community_url = community_url.parts[-1] or community_url.parts[-2]

        await self._state.http.edit_clan(
            self.id64,
            abbreviation=abbreviation,
            headline=headline,
            summary=summary,
            community_url_path=community_url,
            language=language,
            country=country,
            state=state,
            city=city,
            apps=(app.id for app in apps) if apps is not None else None,
        )
        if avatar:
            await self._state.http.update_avatar(
                avatar,
                "group_avatar_image",
                gId=self.id64,
                query={
                    "type": "group_avatar_image",
                    "gId": self.id64,
                    "bgColor": "1b2838",
                },
            )

    async def fetch_members(self) -> AsyncGenerator[PartialUser, None]:
        """An :term:`async iterator` over a clan's member list.

        Note
        ----
        This can be a very slow operation due to the rate limits on this endpoint.
        """
        async for id32 in self._state.http.get_clan_members(self.id64):
            yield self._state.get_partial_user(id32)

    async def join(self) -> None:
        """Joins the clan."""
        await self._state.http.join_clan(self.id64)

    async def leave(self) -> None:
        """Leaves the clan."""
        await self._state.http.leave_clan(self.id64)

    # event/announcement stuff

    async def fetch_event(self, id: int) -> Event[EventType, Self]:
        """Fetch an event from its ID.

        Parameters
        ----------
        id
            The ID of the event.
        """
        try:
            (data,) = await self._state.http.get_clan_events(self.id, [id])
            return await Event(self._state, self, data)
        except ValueError:
            raise ValueError(f"Event {id} not found")

    async def fetch_announcement(self, id: int) -> Announcement[Self]:
        """Fetch an announcement from its ID.

        Parameters
        ----------
        id
            The ID of the announcement.
        """
        data = await self._state.http.get_clan_announcement(self.id, id)
        return await Announcement(self._state, self, data)

    @overload
    async def create_event(
        self,
        name: str,
        content: str,
        *,
        type: Literal[EventType.Game] = ...,
        starts_at: datetime | None = ...,
        app: App,
        server_address: IPAdress | str | None = ...,
        server_password: str | None = ...,
    ) -> Event[Literal[EventType.Game], Self]: ...

    @overload
    async def create_event(
        self,
        name: str,
        content: str,
        *,
        type: BoringEventT = EventType.Other,
        starts_at: datetime | None = None,
    ) -> Event[BoringEventT, Self]: ...

    async def create_event(
        self,
        name: str,
        content: str,
        *,
        type: CreateableEvents = EventType.Other,
        app: App | None = None,
        starts_at: datetime | None = None,
        server_address: IPAdress | str | None = None,
        server_password: str | None = None,
    ) -> Event[CreateableEvents, Self]:
        """Create an event.

        Parameters
        ----------
        name
            The name of the event
        content
            The content for the event.
        type
            The type of the event, defaults to :attr:`EventType.Other`.
        app
            The app that will be played in the event. Required if type is :attr:`EventType.Game`.
        starts_at
            The time the event will start at.
        server_address
            The address of the server that the event will be played on. This is only allowed if ``type`` is
            :attr:`EventType.App`.
        server_password
            The password for the server that the event will be played on. This is only allowed if ``type`` is
            :attr:`EventType.App`.

        Note
        ----
        It is recommended to use a timezone aware datetime for ``start``.

        Returns
        -------
        The created event.
        """
        server_address = IPv4Address(server_address) if server_address is not None else ""

        resp = await self._state.http.create_clan_event(
            self.id64,
            name,
            content,
            f"{type.name}Event",
            str(app.id) if app is not None else "",
            str(server_address),
            server_password or "",
            starts_at,
        )
        soup = BeautifulSoup(resp, HTML_PARSER)
        for element in soup.find_all("div", class_="eventBlockTitle"):
            a = element.a
            if a is not None and a.text == name:  # this is bad?
                _, _, id = a["href"].rpartition("/")
                event = await self.fetch_event(int(id))
                if event.name != name or event.content != content:
                    continue
                self._state.dispatch("event_create", event)
                return cast(Event[CreateableEvents, Self], event)
        raise ValueError

    async def create_announcement(
        self,
        name: str,
        content: str,
        *,
        hidden: bool = False,
    ) -> Announcement[Self]:
        """Create an announcement.

        Parameters
        ----------
        name
            The name of the announcement.
        content
            The content of the announcement.
        hidden
            Whether the announcement should initially be hidden.

        Returns
        -------
        The created announcement.
        """
        await self._state.http.create_clan_announcement(self.id64, name, content, hidden)
        async for entry in self._state.client.user_news(flags=UserNewsType.PostedAnnouncement):
            if entry.target == self._state.user and entry.actor == self:
                announcement = cast("Announcement[Self]", await entry.announcement())
                entry.actor = self  # makes the above line safe
                if announcement.name != name or announcement.content != content:
                    continue
                self._state.dispatch("announcement_create", announcement)
                return announcement

        raise ValueError


class Clan(ChatGroup[ClanMember, ClanChannel, Literal[Type.Clan]], PartialClan):
    """Represents a Steam clan.

    .. container:: operations

        .. describe:: x == y

            Checks if two clans are equal.

        .. describe:: str(x)

            Returns the clan's name.
    """

    __slots__ = (
        "summary",
        "headline",
        "created_at",
        "language",
        "location",
        "member_count",
        "in_game_count",
        "online_count",
        "_is_app_clan",
    )

    # TODO more to implement https://github.com/DoctorMcKay/node-steamcommunity/blob/master/components/groups.js
    # Clan.requesting_membership
    # Clan.respond_to_requesting_membership(*users, approve)
    # Clan.respond_to_all_requesting_membership(approve)

    summary: BBCodeStr | None
    """The summary of the clan."""
    headline: str | None
    """The headline of the clan."""
    created_at: datetime | None
    """The time the clan was created at."""
    member_count: int
    """The amount of users in the clan."""
    online_count: int
    """The amount of users currently online."""
    in_game_count: int
    """The amount of user's currently in game."""
    language: Language
    """The language set for the clan."""
    location: str
    """The location set for the clan."""
    flags: ClanAccountFlags | None
    """The flags"""
    _is_app_clan: bool

    def __init__(self, state: ConnectionState, id: Intable) -> None:
        PartialClan.__init__(self, state, id)
        self._init()
        self.flags = None

    async def _load(self, *, from_proto: bool = False) -> Self:
        community_url = self.community_url
        assert community_url is not None
        async with self._state.http._session.get(community_url) as resp:
            soup = BeautifulSoup(await resp.text(), HTML_PARSER)  # technically we loose proper request handling here
            self._is_app_clan = "games" in resp.url.parts

        if not from_proto:
            assert soup.title is not None
            _, _, self.name = soup.title.text.rpartition(" :: ")
            icon_url = soup.find("link", rel="image_src")
            url = URL(cast(str, icon_url["href"])) if isinstance(icon_url, Tag) else None
            if url:
                self._avatar_sha = bytes.fromhex(url.path.removesuffix("/").removesuffix("_full.jpg"))

        headline = soup.find("div", class_="group_content group_summary")
        headline_h1 = headline.h1 if isinstance(headline, Tag) else None
        self.headline = headline_h1.text if headline_h1 else None
        content = soup.find("meta", property="og:description")
        self.summary = parse_bb_code(cast(str, content["content"])) if isinstance(content, Tag) else None

        if self._is_app_clan:
            for entry in soup.find_all("div", class_="actionItem"):
                if (a := entry.a) is not None:
                    href = a.get("href", "")
                    if match := re.findall(r"store.steampowered.com/app/(\d+)", href):
                        self.app = PartialApp(self._state, id=match[0])
        stats = soup.find("div", class_="grouppage_resp_stats")
        if stats is None:
            return self
        assert isinstance(stats, Tag)
        for stat in stats.find_all("div", class_="groupstat"):
            if "Founded" in stat.text:
                text = stat.text.split("Founded")[1].strip()
                if ", " not in stat.text:
                    text = f"{text}, {DateTime.now().year}"
                self.created_at = DateTime.parse_steam_date(text)
            if "Language" in stat.text:
                self.language = stat.text.split("Language")[1].strip()
            if "Location" in stat.text:
                self.location = stat.text.split("Location")[1].strip()

        for count in stats.find_all("div", class_="membercount"):
            if "MEMBERS" in count.text:
                self.member_count = int(count.text.split("MEMBERS")[0].strip().replace(",", ""))
            if "IN-GAME" in count.text:
                self.in_game_count = int(count.text.split("IN-GAME")[0].strip().replace(",", ""))
            if "ONLINE" in count.text:
                self.online_count = int(count.text.split("ONLINE")[0].strip().replace(",", ""))

        if not from_proto:
            self._officers: list[ID32] = []
            self._mods: list[ID32] = []
            is_admins = None
            for fields in soup.find_all("div", class_="membergrid"):
                for field in fields.find_all("div"):
                    if "Administrators" in field.text:
                        is_admins = True
                        continue
                    if "Moderators" in field.text:
                        is_admins = False
                        continue
                    if "Members" in field.text:
                        break

                    try:
                        id = ID32(int(field["data-miniprofile"]))
                    except KeyError:
                        continue
                    else:
                        if is_admins is None:
                            continue
                        if is_admins:
                            self._officers.append(id)
                        else:
                            self._mods.append(id)

            self._officers_and_mods = await self._state._maybe_users(map(parse_id64, self._officers + self._mods))

        return self

    @classmethod
    async def _from_proto(  # type: ignore
        cls,
        state: ConnectionState,
        proto: chat.GetChatRoomGroupSummaryResponse,
        *,
        maybe_chunk: bool = True,
    ) -> Self:
        self = await super()._from_proto(state, proto, id=ID32(proto.clanid), maybe_chunk=maybe_chunk)
        return await self._load(from_proto=True)

    async def chunk(self) -> Sequence[ClanMember]:
        if self.chunked:
            return self.members

        self._members = dict.fromkeys(self._partial_members)  # type: ignore
        if self.flags & ClanAccountFlags.Large > 0 if self.flags is not None else len(self._partial_members) <= 100:
            for user, member in zip(
                await self._state._maybe_users(parse_id64(id, type=Type.Individual) for id in self._partial_members),
                self._partial_members.values(),
            ):
                self._members[user.id] = ClanMember(self._state, self, user, member)

            return await super().chunk()

        # these actually need fetching
        view_id = self._state.chat_group_to_view_id[self._id]
        users = {
            user.id: user
            for users in await asyncio.gather(
                *(
                    self._state.fetch_chat_group_members(
                        self._id,
                        view_id,
                        client_change_number
                        + 1,  # steam doesn't send responses if they're 0 (TODO this might be a betterproto bug)
                        start + 1,
                        stop,
                    )
                    for client_change_number, (start, stop) in enumerate(
                        utils._int_chunks(len(self._partial_members), 100)
                    )
                )
            )
            for user in users
        }
        for id, member in self._partial_members.items():
            try:
                user = users[id]
            except KeyError:
                user = await self._state._maybe_user(parse_id64(id))
            self._members[user.id] = ClanMember(self._state, self, user, member)
        return await super().chunk()

    def _get_partial_member(self, id: ID32, /) -> PartialMember:
        try:
            return PartialMember(self._state, clan=self, member=self._partial_members[id])
        except KeyError:
            member = self._partial_members[id] = chat.Member(id, state=chat.EChatRoomJoinState.NONE)
            return PartialMember(self._state, clan=self, member=member)

    def is_ogg(self) -> bool:
        """Whether this clan is an official game group."""
        if self.flags is not None:
            return self.flags & ClanAccountFlags.OGG > 0
        return self._is_app_clan

    async def edit(
        self,
        *,
        abbreviation: str | None = None,
        headline: str | None = None,
        summary: str | None = None,
        community_url: URL | str | None = None,
        language: Language | None = None,
        country: str | None = None,
        state: str | None = None,
        city: str | None = None,
        apps: Iterable[App] | None = None,
        name: str | None = None,
        tagline: str | None = None,
        avatar: Media | None = None,
    ) -> None:
        """Edits the clan.

        Parameters
        ----------
        name
            The new name for the clan's chat group. **Does not change the clan's name.**
        tagline
            The new tagline for the clan.
        avatar
            The new avatar for the clan.

        See Also
        --------
        :meth:`PartialClan.edit` for other parameters.
        """
        await PartialClan.edit(
            self,
            abbreviation=abbreviation,
            headline=headline,
            summary=summary,
            community_url=community_url,
            language=language,
            country=country,
            state=state,
            city=city,
            apps=apps,
            avatar=avatar,
        )
        await super().edit(name=name, tagline=tagline)

    async def fetch_members(self) -> AsyncGenerator[ClanMember | PartialMember, None]:
        async for id32 in self._state.http.get_clan_members(self.id64):
            yield self._maybe_member(id32)

    async def join(self, *, invite_code: str | None = None) -> None:
        await self._state.http.join_clan(self.id64)
        await super().join(invite_code=invite_code)

    async def leave(self) -> None:
        await super().leave()
        await self._state.http.leave_clan(self.id64)

    async def events(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AsyncGenerator[Event[EventType, Self], None]:
        """An :term:`asynchronous iterator` over a clan's :class:`steam.Event`\\s.

        Examples
        --------

        Usage:

        .. code:: python

            async for event in clan.events(limit=10):
                print(event.author, "made an event", event.name, "starting at", event.starts_at)

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of events to search through. Default is ``100``. Setting this to ``None`` will fetch all
            the clan's events, but this will be a very slow operation.
        before
            A time to search for events before.
        after
            A time to search for events after.

            Warning
            -------
            If this is ``None`` and :attr:`created_at` is ``None``, this has to fetch events all the way back until 2007

        Yields
        ---------
        :class:`~steam.Event`
        """

        after = after or self.created_at or datetime(2007, 7, 1, tzinfo=timezone.utc)
        # uh-oh I really hope you have a created_at cause I have no way of telling if this request is done
        # date from https://www.ign.com/articles/2007/08/06/steam-community-beta-opens (public release of steamcommunity)
        before = before or DateTime.now()
        start_at_month = after.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        stop_at_month = after.replace(day=1, hour=0, minute=0, second=0, microsecond=0)

        dates: list[date] = []
        start_month = start_at_month.month
        stop_month = 13

        for year in range(start_at_month.year, stop_at_month.year + 1):
            if year == stop_at_month.year:
                stop_month = stop_at_month.month + 1
            dates += (date(year, month, 1) for month in range(start_month, stop_month))
            start_month = 1

        ids = [
            id
            for date_chunk in utils.as_chunks(dates, 12)
            for ids in await asyncio.gather(
                *(self._state.http.get_clan_events_for(self.id64, date) for date in date_chunk)
            )
            for id in ids
        ]

        yielded = 0

        for id_chunk in utils.as_chunks(ids, 15):
            events: list[Event[EventType, Self]] = []
            for event_ in await self._state.http.get_clan_events(self.id, id_chunk):
                event = Event(self._state, self, event_)
                if not after < event.starts_at < before:
                    break
                events.append(event)

            authors = utils.as_chunks(
                await self._state._maybe_users(
                    itertools.chain.from_iterable(
                        (
                            e.author.id64,
                            e.last_edited_by.id64 if e.last_edited_by is not None else ID64(0),
                        )
                        for e in events
                    )
                ),
                2,
            )
            for event in events:
                if limit is not None and yielded >= limit:
                    return
                event.author, event.last_edited_by = next(authors)
                yield event
                yielded += 1

    async def announcements(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
        # hidden: bool = False,
    ) -> AsyncGenerator[Announcement[Self], None]:
        """An :term:`asynchronous iterator` over a clan's :class:`steam.Announcement`\\s.

        Examples
        --------

        Usage:

        .. code:: python

            async for announcement in clan.announcements(limit=10):
                print(announcement.author, "made an announcement", announcement.name, "at", announcement.created_at)

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of announcements to search through. Default is ``100``. Setting this to ``None`` will
            fetch all of the clan's announcements, but this will be a very slow operation.
        before
            A time to search for announcements before.
        after
            A time to search for announcements after.

        Yields
        ---------
        :class:`~steam.Announcement`
        """
        after = after or UNIX_EPOCH
        before = before or DateTime.now()

        ids = await self._state.http.get_clan_announcement_ids(
            self.id64
        )  # TODO make this use the calendar? does that work for announcements

        if not ids:
            return

        announcements: list[Announcement[Self]] = []
        for announcement_ in await asyncio.gather(*(self._state.http.get_clan_announcement(self.id, id) for id in ids)):
            announcement = Announcement(self._state, self, announcement_)
            if not after < announcement.starts_at < before:
                break
            announcements.append(announcement)

        for yielded, (announcement, (author, last_edited_by)) in enumerate(
            zip(
                announcements,
                utils.as_chunks(
                    await self._state._maybe_users(
                        itertools.chain.from_iterable(
                            (
                                a.author.id64,
                                a.last_edited_by.id64 if a.last_edited_by is not None else ID64(0),
                            )
                            for a in announcements
                        )
                    ),
                    2,
                ),
            )
        ):
            if limit is not None and yielded >= limit:
                return
            announcement.author = author
            announcement.last_edited_by = last_edited_by if last_edited_by.id != 0 else None
            yield announcement



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: gcsdk_gcmessages.proto
# plugin: python-betterproto

from dataclasses import dataclass

import betterproto

from ....protobufs.msg import GCProtobufMessage
from ..enums import EMsg


class PartnerAccountType(betterproto.Enum):
    NONE = 0
    PerfectWorld = 1
    Nexon = 2


class GCConnectionStatus(betterproto.Enum):
    HaveSession = 0
    GcGoingDown = 1
    NoSession = 2
    NoSessionInLogonQueue = 3
    NoSteam = 4
    Suspended = 5


@dataclass(eq=False, repr=False)
class IDOwner(betterproto.Message):
    type: int = betterproto.uint32_field(1)
    id: int = betterproto.uint64_field(2)


class SOUpdate(GCProtobufMessage, msg=EMsg.SOUpdate):
    owner: int = betterproto.fixed64_field(1)
    type_id: int = betterproto.int32_field(2)
    object_data: bytes = betterproto.bytes_field(3)
    version: int = betterproto.fixed64_field(4)
    owner_soid: "IDOwner" = betterproto.message_field(5)
    service_id: int = betterproto.uint32_field(6)


class SOCreate(GCProtobufMessage, msg=EMsg.SOCreate):
    owner: int = betterproto.fixed64_field(1)
    type_id: int = betterproto.int32_field(2)
    object_data: bytes = betterproto.bytes_field(3)
    version: int = betterproto.fixed64_field(4)
    owner_soid: "IDOwner" = betterproto.message_field(5)
    service_id: int = betterproto.uint32_field(6)


class SODestroy(GCProtobufMessage, msg=EMsg.SODestroy):
    owner: int = betterproto.fixed64_field(1)
    type_id: int = betterproto.int32_field(2)
    object_data: bytes = betterproto.bytes_field(3)
    version: int = betterproto.fixed64_field(4)
    owner_soid: "IDOwner" = betterproto.message_field(5)
    service_id: int = betterproto.uint32_field(6)


class GameServerResetIdentityResponse(GCProtobufMessage, msg=EMsg.GameServerResetIdentityResponse):
    owner: int = betterproto.fixed64_field(1)
    type_id: int = betterproto.int32_field(2)
    object_data: bytes = betterproto.bytes_field(3)
    version: int = betterproto.fixed64_field(4)
    owner_soid: "IDOwner" = betterproto.message_field(5)
    service_id: int = betterproto.uint32_field(6)


class MultipleObjects(GCProtobufMessage, msg=EMsg.SOUpdateMultiple):
    owner: int = betterproto.fixed64_field(1)
    objects: "list[MultipleObjectsSingleObject]" = betterproto.message_field(2)
    version: int = betterproto.fixed64_field(3)
    owner_soid: "IDOwner" = betterproto.message_field(6)
    service_id: int = betterproto.uint32_field(7)


@dataclass(eq=False, repr=False)
class MultipleObjectsSingleObject(betterproto.Message):
    type_id: int = betterproto.int32_field(1)
    object_data: bytes = betterproto.bytes_field(2)


class CacheSubscribed(GCProtobufMessage, msg=EMsg.SOCacheSubscribed):
    owner: int = betterproto.fixed64_field(1)
    objects: "list[CacheSubscribedSubscribedType]" = betterproto.message_field(2)
    version: int = betterproto.fixed64_field(3)
    owner_soid: "IDOwner" = betterproto.message_field(4)
    service_id: int = betterproto.uint32_field(5)
    service_list: list[int] = betterproto.uint32_field(6)
    sync_version: int = betterproto.fixed64_field(7)


@dataclass(eq=False, repr=False)
class CacheSubscribedSubscribedType(betterproto.Message):
    type_id: int = betterproto.int32_field(1)
    object_data: list[bytes] = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class CacheSubscribedUpToDate(betterproto.Message):
    version: int = betterproto.fixed64_field(1)
    owner_soid: "IDOwner" = betterproto.message_field(2)
    service_id: int = betterproto.uint32_field(3)
    service_list: list[int] = betterproto.uint32_field(4)
    sync_version: int = betterproto.fixed64_field(5)


@dataclass(eq=False, repr=False)
class CacheUnsubscribed(betterproto.Message):
    owner: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CacheSubscriptionCheck(betterproto.Message):
    owner: int = betterproto.fixed64_field(1)
    version: int = betterproto.fixed64_field(2)
    owner_soid: "IDOwner" = betterproto.message_field(3)
    service_id: int = betterproto.uint32_field(4)
    service_list: list[int] = betterproto.uint32_field(5)
    sync_version: int = betterproto.fixed64_field(6)


class CacheSubscriptionRefresh(GCProtobufMessage, msg=EMsg.SOCacheSubscriptionRefresh):
    owner: int = betterproto.fixed64_field(1)
    owner_soid: "IDOwner" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class CacheVersion(betterproto.Message):
    version: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CMultiplexMessage(betterproto.Message):
    msgtype: int = betterproto.uint32_field(1)
    payload: bytes = betterproto.bytes_field(2)
    steamids: list[float] = betterproto.fixed64_field(3)


@dataclass(eq=False, repr=False)
class MasterAck(betterproto.Message):
    dir_index: int = betterproto.uint32_field(1)
    machine_name: str = betterproto.string_field(3)
    process_name: str = betterproto.string_field(4)
    type_instances: list[int] = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class MasterAckResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class MasterStartupComplete(betterproto.Message):
    gc_info: "list[MasterStartupCompleteGCInfo]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class MasterStartupCompleteGCInfo(betterproto.Message):
    dir_index: int = betterproto.uint32_field(1)
    machine_name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class Routed(betterproto.Message):
    msg_type: int = betterproto.uint32_field(1)
    sender_id: int = betterproto.fixed64_field(2)
    net_message: bytes = betterproto.bytes_field(3)


@dataclass(eq=False, repr=False)
class RoutedReply(betterproto.Message):
    msg_type: int = betterproto.uint32_field(1)
    net_message: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class UpdateSubGCSessionInfo(betterproto.Message):
    updates: "list[UpdateSubGCSessionInfoUpdate]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class UpdateSubGCSessionInfoUpdate(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    ip: float = betterproto.fixed32_field(2)
    trusted: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class RequestSubGCSessionInfo(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class RequestSubGCSessionInfoResponse(betterproto.Message):
    ip: float = betterproto.fixed32_field(1)
    trusted: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class ToGCIncrementRecruitmentLevel(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CacheHaveVersion(betterproto.Message):
    soid: "IDOwner" = betterproto.message_field(1)
    version: int = betterproto.fixed64_field(2)
    service_id: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class ConnectionStatus(betterproto.Message):
    status: "GCConnectionStatus" = betterproto.enum_field(1)
    client_session_need: int = betterproto.uint32_field(2)
    queue_position: int = betterproto.int32_field(3)
    queue_size: int = betterproto.int32_field(4)
    wait_seconds: int = betterproto.int32_field(5)
    estimated_wait_seconds_remaining: int = betterproto.int32_field(6)


@dataclass(eq=False, repr=False)
class CacheSubscribe(betterproto.Message):
    subscriber: int = betterproto.fixed64_field(1)
    subscribe_to: int = betterproto.fixed64_field(2)
    sync_version: int = betterproto.fixed64_field(3)
    have_versions: "list[CacheSubscribeHaveVersions]" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class CacheSubscribeHaveVersions(betterproto.Message):
    service_id: int = betterproto.uint32_field(1)
    version: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CacheUnsubscribe(betterproto.Message):
    subscriber: int = betterproto.fixed64_field(1)
    unsubscribe_from: int = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class ClientPing(betterproto.Message):
    pass



# pyright: reportIncompatibleVariableOverride = false
"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import dataclasses
import functools
import importlib
import logging
import sys
from dataclasses import dataclass
from inspect import get_annotations
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, ClassVar, Final, TypeVar

import betterproto
from typing_extensions import Self, dataclass_transform

from .._const import MISSING, SET_PROTO_BIT, WRITE_U32
from ..enums import IntEnum, Result
from ..utils import classproperty
from . import GC_PROTOBUFS, PROTOBUFS, UMS
from .emsg import *
from .headers import *
from .headers import MessageHeader
from .struct_messages import StructMessage

if TYPE_CHECKING:
    from collections.abc import Callable

    from ..types.id import AppID

log = logging.getLogger(__name__)

__all__ = (
    "NoMsg",
    "Message",
    "ProtobufMessage",
    "UnifiedMessage",
    "GCMessage",
    "GCProtobufMessage",
    "REQUEST_EMSGS",
    "RESPONSE_EMSGS",
    "SERVICE_EMSGS",
)

TT = TypeVar("TT", bound=type[object])


def load_annotations(cls: TT) -> TT:
    """Needed so dataclasses don't break when evaluating ForwardRefs"""
    cls.__annotations__ = get_annotations(cls, eval_str=True)
    return cls


class NoMsg(IntEnum):
    NONE = -2


@load_annotations
class MessageBase:
    __slots__ = ()

    MSG: ClassVar[IntEnum] = NoMsg.NONE
    header: MessageHeader | ProtobufMessageHeader | GCMessageHeader

    def __init_subclass__(cls, **kwargs: Any) -> object:
        if cls.__module__ == __name__:
            return load_annotations(cls)

        if keys := cls.__annotations__.keys() & {"header", "result", "parse", "MSG", "UM_NAME", "APP_ID"}:
            raise TypeError(f"{', '.join(keys)} is/are a reserved field name(s)")

        return dataclass(**({"repr": False, "eq": False} | kwargs))(cls)

    @property
    def result(self) -> Result:
        return Result.try_value(getattr(self, "eresult", 0) or self.header.eresult)

    if not TYPE_CHECKING:
        # shim to make betterproto's repr work
        @classproperty
        def __dataclass_fields__(cls: type[Self], value: MappingProxyType[Any, Any] = MappingProxyType({})) -> Any:
            return value

        @classproperty
        def __dataclass_params__(cls: type[Self], value=dataclasses.make_dataclass("", ()).__dataclass_params__) -> Any:
            return value


@load_annotations
class MessageMessageBase:
    __slots__ = ()

    MSG: ClassVar[EMsg]

    def __init_subclass__(cls, msg: EMsg = MISSING, **kwargs: bool) -> object:
        if cls.__module__ == __name__:
            return super().__init_subclass__(**kwargs)

        if msg is MISSING:
            # should be a dataclass with added slots (it recreates the class so the kwarg is removed)
            try:
                msg = cls.MSG
            except AttributeError:
                raise TypeError(
                    "MessageMessageBase.__init_subclass__() missing 1 required positional argument: 'msg'"
                ) from None
        cls.MSG = msg
        PROTOBUFS[msg] = cls  # type: ignore
        return super().__init_subclass__(**kwargs)


class NotProtobufWrapped(MessageBase, StructMessage):
    __slots__ = ()
    header: MessageHeader | GCMessageHeader

    def __init_subclass__(cls, **kwargs: Any) -> None:
        if cls.MSG is not NoMsg.NONE:

            def __bytes__(
                self: Self,
                starting_header: bytes = WRITE_U32(cls.MSG),
                bytes_method: Callable[[Self], bytes] = (
                    StructMessage.__bytes__ if cls.__bytes__ is NotProtobufWrapped.__bytes__ else cls.__bytes__
                ),
                /,
            ) -> bytes:
                return starting_header + bytes(self.header) + bytes_method(self)

            cls.__bytes__ = __bytes__
        super().__init_subclass__(**kwargs)

    def __bytes__(self, /) -> bytes:
        raise NotImplementedError


@dataclass_transform()
class Message(MessageMessageBase, NotProtobufWrapped):
    def __post_init__(self) -> None:
        self.header: MessageHeader = MessageHeader()

    __init__ = __post_init__

    def __init_subclass__(cls, msg: EMsg = MISSING, **kwargs: bool) -> object:
        return super().__init_subclass__(msg, repr=True, eq=True, **kwargs)

    def parse(  # type: ignore
        self,
        data: bytes,
        msg: int,
        /,
    ) -> Self:
        try:
            new_class: type[Self] = PROTOBUFS[msg]  # type: ignore
        except KeyError:
            log.debug("Received an unknown %r (%s)", EMsg(msg), data)
            return self

        self.header.parse(data)
        self.__class__ = new_class
        return self.parse(data[self.header.length :])  # type: ignore  # this will then call the new parse method


class ProtobufWrappedMessage(MessageBase, betterproto.Message):
    __slots__ = ()
    header: ProtobufMessageHeader

    def __post_init__(self) -> None:
        self.header = ProtobufMessageHeader(job_name_target=getattr(self.__class__, "UM_NAME", ""))
        return super().__post_init__()

    __init__ = __post_init__

    def __bytes__(self) -> bytes:
        return WRITE_U32(SET_PROTO_BIT(self.__class__.MSG)) + bytes(self.header) + betterproto.Message.__bytes__(self)


REQUEST_EMSGS: Final = frozenset(
    {EMsg.ServiceMethod, EMsg.ServiceMethodCallFromClient, EMsg.ServiceMethodCallFromClientNonAuthed}
)
RESPONSE_EMSGS: Final = frozenset({EMsg.ServiceMethodResponse, EMsg.ServiceMethodSendToClient})
SERVICE_EMSGS: Final = frozenset({*REQUEST_EMSGS, *RESPONSE_EMSGS})


@dataclass_transform()
class ProtobufMessage(MessageMessageBase, ProtobufWrappedMessage):
    def parse(  # type: ignore
        self,
        data: bytes,
        msg: int = MISSING,
        /,
    ) -> Self:
        if msg is MISSING:  # case CMsgMulti().parse(data)
            return betterproto.Message.parse(self, data)  # type: ignore  # pyright's dumb

        self.header.parse(data)
        if msg in SERVICE_EMSGS:
            try:
                new_class = UMS[self.header.job_name_target][msg in RESPONSE_EMSGS]  # type: ignore
            except KeyError:
                log.debug("Received an unknown UM %r (%s)", self.header.job_name_target, data)
                return self
        else:
            try:
                new_class: type[Self] = PROTOBUFS[msg]  # type: ignore  # save the extra lookup when casting to EMsg
            except KeyError:
                log.debug("Received an unknown %r (%s)", EMsg(msg), data)
                return self

        try:
            self.__class__ = new_class
        except TypeError:
            # setting to MISSING
            log.info("Received an unknown %r %r (%s)", EMsg(msg), data, self.header.job_name_target)
            return self
        return betterproto.Message.parse(self, data[self.header.length :])  # type: ignore  # pyright's dumb


@dataclass_transform()
class UnifiedMessage(ProtobufMessage):
    UM_NAME: ClassVar[str]

    def __init_subclass__(cls, um_name: str, msg: EMsg = MISSING) -> None:
        cls.UM_NAME = sys.intern(f"{um_name}#1")

        if msg is MISSING:
            msg = (
                EMsg.ServiceMethodCallFromClient
                if cls.__name__.endswith(("Request", "Notification"))
                else EMsg.ServiceMethodSendToClient
            )

        try:
            um = UMS[cls.UM_NAME]
        except KeyError:
            UMS[cls.UM_NAME] = um = (MISSING, MISSING)  # type: ignore

        UMS[cls.UM_NAME] = (cls, um[1]) if msg in REQUEST_EMSGS else (um[0], cls)  # type: ignore

        super().__init_subclass__(msg)


@functools.lru_cache
def get_app_id(module: str) -> AppID:
    gc_protobuf_locations, _, _ = module.rpartition(".")
    protobufs_module = importlib.import_module(gc_protobuf_locations)
    return protobufs_module.APP_ID


class GCMessageBase:
    __slots__ = ()

    if TYPE_CHECKING:
        APP_ID: ClassVar[AppID]
        MSG: ClassVar[IntEnum]

    def __init_subclass__(cls, msg: IntEnum = MISSING) -> None:
        if cls.__module__ == __name__:
            return super().__init_subclass__()

        if msg is MISSING:
            raise TypeError("GCMessageBase.__init_subclass__() missing 1 required positional argument: 'msg'")

        cls.APP_ID = get_app_id(cls.__module__)
        try:
            GC_PROTOBUFS[cls.APP_ID][msg] = cls  # type: ignore
        except KeyError:
            GC_PROTOBUFS[cls.APP_ID] = {msg: cls}  # type: ignore
        cls.MSG = msg
        super().__init_subclass__()


@dataclass_transform()
class GCMessage(GCMessageBase, NotProtobufWrapped):
    """A wrapper around received GC messages, mainly for extensions."""

    def __post_init__(self) -> None:
        self.header = GCMessageHeader()

    __init__ = __post_init__

    def parse(  # type: ignore
        self,
        data: bytes,
        msg: int,
        app_id: AppID,
        /,
    ) -> Self:
        try:
            new_class: type[Self] = GC_PROTOBUFS[app_id][msg]  # type: ignore  # save the extra lookup when casting to IntEnum
        except KeyError:
            log.debug("Received an unknown EMsg %d for %d (%s)", msg, app_id, data)
            return self

        self.header.parse(data)
        self.__class__ = new_class
        return self.parse(data[self.header.length :])  # type: ignore


@dataclass_transform()
class GCProtobufMessage(GCMessageBase, ProtobufWrappedMessage):
    """A wrapper around received GC protobuf messages, mainly for extensions."""

    def parse(  # type: ignore
        self,
        data: bytes,
        msg: int = MISSING,
        app_id: AppID = MISSING,
        /,
    ) -> Self:
        if msg is MISSING:  # case CMsgMulti().parse(data)
            return betterproto.Message.parse(self, data)  # type: ignore  # pyright's dumb

        try:
            new_class: type[Self] = GC_PROTOBUFS[app_id][msg]  # type: ignore  # save the extra lookup when casting to IntEnum
        except KeyError:
            log.debug("Received an unknown EMsg %d for %d (%s)", msg, app_id, data)
            return self

        self.header.parse(data)
        self.__class__ = new_class
        return betterproto.Message.parse(self, data[self.header.length :])  # type: ignore  # pyright's dumb



"""
Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE

Contains large portions of:
https://github.com/Rapptz/discord.py/blob/master/discord/client.py
The appropriate license is in LICENSE
"""

from __future__ import annotations

import asyncio
import inspect
import logging
import math
import random
import sys
import time
import traceback
from collections.abc import AsyncGenerator, Callable, Collection, Coroutine, Iterable, Sequence
from contextlib import nullcontext
from ipaddress import IPv4Address
from typing import (
    TYPE_CHECKING,
    Any,
    Concatenate,
    Literal,
    ParamSpec,
    TypeAlias,
    TypedDict,
    TypeVar,
    cast,
    final,
    overload,
)

from yarl import URL as URL_

from . import utils
from ._const import DOCS_BUILDING, STATE, UNIX_EPOCH, URL, TaskGroup, timeout
from .achievement import UserNewsAchievement
from .app import App, AppListApp, AuthenticationTicket, FetchedApp, PartialApp
from .bundle import Bundle, FetchedBundle, PartialBundle
from .chat import ChatGroup
from .enums import *
from .errors import WSException
from .game_server import GameServer, Query
from .gateway import *
from .guard import get_authentication_code
from .http import HTTPClient
from .id import _ID64_TO_ID32
from .models import CDNAsset, PriceOverview, Wallet, return_true
from .package import FetchedPackage, License, Package, PartialPackage
from .protobufs import store
from .state import ConnectionState
from .store import AppStoreItem, BundleStoreItem, PackageStoreItem, TransactionReceipt
from .types.id import ID64, AppID, BundleID, PackageID, PublishedFileID, TradeOfferID
from .user_news import UserNews
from .utils import DateTime, TradeURLInfo

if TYPE_CHECKING:
    import datetime
    from ssl import SSLContext

    import aiohttp
    from typing_extensions import Self, Unpack

    import steam
    from steam.ext import commands

    from .abc import Message
    from .clan import Clan
    from .comment import Comment
    from .event import Announcement, Event
    from .ext.commands.bot import Bot
    from .friend import Friend
    from .group import Group
    from .invite import AppInvite, ClanInvite, GroupInvite, UserInvite
    from .manifest import AppInfo, PackageInfo
    from .media import Media
    from .post import Post
    from .published_file import PublishedFile
    from .reaction import ClientEffect, ClientEmoticon, ClientSticker, MessageReaction
    from .trade import Asset, Item, MovedItem, TradeOffer
    from .types.http import IPAdress
    from .types.user import IndividualID
    from .user import ClientUser, User


__all__ = ("Client",)

log = logging.getLogger(__name__)

CoroFunc: TypeAlias = Callable[..., Coroutine[Any, Any, Any]]
F = TypeVar("F", bound=CoroFunc)
P = ParamSpec("P")


class ClientKwargs(TypedDict, total=False):
    proxy: str | None
    proxy_auth: aiohttp.BasicAuth | None
    connector: aiohttp.BaseConnector | None
    intents: Intents
    max_messages: int | None
    app: App | None
    apps: list[App]
    state: PersonaState
    ui_mode: UIMode
    flags: PersonaStateFlag
    force_kick: bool
    language: Language
    auto_chunk_chat_groups: bool
    ssl: SSLContext | Literal[False] | aiohttp.Fingerprint


class Client:
    """Represents a client connection that connects to Steam. This class is used to interact with the Steam API and CMs.

    .. container:: operations

        .. describe:: async with x

            Initialises the client and closes it when the context is exited.

    Parameters
    ----------
    proxy
        A proxy URL to use for requests.
    proxy_auth
        The proxy authentication to use with requests.
    connector
        The connector to use with the :class:`aiohttp.ClientSession`.
    intents
        The intents you wish to start the client with.
    max_messages
        The maximum number of messages to store in the internal cache, default is 1000.
    app
        An app to set your status as on connect. This will take precedence over any apps set using ``apps`` as your
        current status.
    apps
        A list of apps to set your status to on connect.
    state
        The state to show your account as on connect.

        Note
        ----
        Setting your status to :attr:`~steam.PersonaState.Offline`, will stop you receiving persona state updates
        and by extension :meth:`on_user_update` will stop being dispatched.
    ui_mode
        The UI mode to set your status to on connect.
    flags
        Flags to set your persona state to.
    force_kick
        Whether to forcefully kick any other playing sessions on connect. Defaults to ``False``.
    language
        The language to use when interacting with the API.
    auto_chunk_chat_groups
        Whether to automatically call chunk on clans and groups filling :attr:`ChatGroup.members`. Setting this to
        ``True`` isn't recommend unless you have a good internet connection and good hardware.
    ssl
        Any ``ssl`` parameters to pass to the underlying :class:`~aiohttp.ClientSession`.

        .. versionadded:: 1.0.1
    """

    def __init__(self, **options: Unpack[ClientKwargs]):
        self.http = HTTPClient(client=self, **options)
        self._state = self._get_state(**options)
        self.ws: SteamWebSocket | None = None

        self.username: str | None = None
        self.password: str | None = None
        self.shared_secret: str | None = None
        self.identity_secret: str | None = None

        self._closed = True
        self._listeners: dict[str, list[tuple[asyncio.Future[Any], Callable[..., bool]]]] = {}
        self._ready = asyncio.Event()
        self._aentered = False

    def _get_state(self, **options: Any) -> ConnectionState:
        return ConnectionState(client=self, **options)

    @property
    def user(self) -> ClientUser:
        """Represents the connected client. ``None`` if not logged in."""
        return self.http.user

    @property
    def users(self) -> Sequence[User]:
        """A read-only list of all the users the connected client can see."""
        return self._state.users

    @property
    def trades(self) -> Sequence[TradeOffer[Asset[User], Asset[ClientUser], User]]:
        """A read-only list of all the trades the connected client can see."""
        return self._state.trades

    @property
    def groups(self) -> Sequence[Group]:
        """A read-only list of all the groups the connected client is in."""
        return self._state.groups

    @property
    def messages(self) -> Sequence[Message]:
        """A read-only list of all the messages the client has."""
        return self._state._messages

    @property
    def clans(self) -> Sequence[Clan]:
        """A read-only list of all the clans the connected client is in."""
        return self._state.clans

    @property
    def licenses(self) -> Sequence[License]:
        """A read-only list of licenses the client has access to."""
        return list(self._state.licenses.values())

    @property
    def emoticons(self) -> Sequence[ClientEmoticon]:
        """A read-only list of all the emoticons the client has."""
        return self._state.emoticons

    @property
    def stickers(self) -> Sequence[ClientSticker]:
        """A read-only list of all the stickers the client has."""
        return self._state.stickers

    @property
    def effects(self) -> Sequence[ClientEffect]:
        """A read-only list of all the effects the client has."""
        return self._state.effects

    @property
    def latency(self) -> float:
        """Measures latency between a heartbeat send and the heartbeat interval in seconds."""
        return float("nan") if self.ws is None else self.ws.latency

    @property
    def refresh_token(self) -> str | None:
        """The refresh token for the logged in account, can be used to login."""
        return None if self.ws is None else self.ws.refresh_token

    @property
    def wallet(self) -> Wallet:
        """The wallet info for the logged in account."""
        return self._state.wallet

    async def code(self) -> str:
        """Get the current steam guard code.

        Warning
        -------
        This will read from :attr:`sys.stdin` if no `shared_secret` was passed to :meth:`login`.
        """
        if self.shared_secret:
            return get_authentication_code(self.shared_secret)
        code = await utils.ainput(">>> ")
        return code.strip()

    def is_ready(self) -> bool:
        """Specifies if the client's internal cache is ready for use."""
        return self._ready.is_set()

    def is_closed(self) -> bool:
        """Indicates if connection is closed to the WebSocket."""
        return self._closed

    def event(self, coro: F) -> F:
        """A decorator that registers an event to listen to.

        The events must be a :ref:`coroutine <coroutine>`, if not, :exc:`TypeError` is raised.

        Usage:

        .. code:: python

            @client.event
            async def on_ready():
                print("Ready!")

        Raises
        ------
        :exc:`TypeError`
            The function passed is not a coroutine.
        """

        if not inspect.iscoroutinefunction(coro):
            raise TypeError(f"Registered events must be coroutine functions, {coro.__name__} is {type(coro).__name__}")

        setattr(self, coro.__name__, coro)
        log.debug("%s has been registered as an event", coro.__name__)
        return coro

    async def _run_event(self, coro: CoroFunc, event_name: str, *args: Any, **kwargs: Any) -> None:
        try:
            await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception as exc:
            try:
                await self.on_error(event_name, exc, *args, **kwargs)
            except asyncio.CancelledError:
                pass

    def _schedule_event(self, coro: CoroFunc, event_name: str, *args: Any, **kwargs: Any) -> asyncio.Task[None]:
        return self._tg.create_task(
            self._run_event(coro, event_name, *args, **kwargs), name=f"steam.py task: {event_name}"
        )

    def dispatch(self, event: str, /, *args: Any, **kwargs: Any) -> None:
        log.debug("Dispatching event %s", event)
        method = f"on_{event}"

        # remove the dispatched listener
        if listeners := self._listeners.get(event):
            removed: list[int] = []
            for idx, (future, condition) in enumerate(listeners):
                if future.cancelled():
                    removed.append(idx)
                    continue

                try:
                    result = condition(*args)
                except Exception as exc:
                    future.set_exception(exc)
                    removed.append(idx)
                else:
                    if result:
                        if not args:
                            future.set_result(None)
                        elif len(args) == 1:
                            future.set_result(args[0])
                        else:
                            future.set_result(args)
                        removed.append(idx)

            if len(removed) == len(listeners):
                del self._listeners[event]
            else:
                for idx in reversed(removed):
                    del listeners[idx]

        # schedule the event (if possible)
        try:
            coro = getattr(self, method)
        except AttributeError:
            pass
        else:
            self._schedule_event(coro, method, *args, **kwargs)

    async def _handle_ready(self) -> None:
        self._ready.set()
        self.dispatch("ready")

    async def __aenter__(self) -> Self:
        self._tg = TaskGroup()
        self._aentered = True
        await self._tg.__aenter__()
        return self

    async def __aexit__(self, *args: Any) -> None:
        if not self.is_closed():
            await self.close()
        if self._aentered:
            await self._tg.__aexit__(*args)

    @overload
    @final
    def run(
        self,
        username: str,
        password: str,
        *,
        shared_secret: str = ...,
        identity_secret: str = ...,
        debug: bool = ...,
    ) -> object: ...

    @overload
    @final
    def run(
        self,
        *,
        refresh_token: str,
        shared_secret: str = ...,
        identity_secret: str = ...,
        debug: bool = ...,
    ) -> object: ...

    @final
    def run(self, *args: Any, debug: bool = False, **kwargs: Any) -> object:
        """A blocking method to start and run the client.

        Shorthand for:

        .. code:: python

            async def main():
                async with client:
                    await client.login(...)


            asyncio.run(main())

        It is not recommended to subclass this method, it is normally favourable to subclass :meth:`login` as it is a
        :ref:`coroutine <coroutine>`.

        Note
        ----
        This takes the same arguments as :meth:`login`.
        """

        async def runner() -> None:
            async with self:
                await self.login(*args, **kwargs)

        try:
            asyncio.run(runner(), debug=debug)
        except KeyboardInterrupt:
            log.info("Closing the event loop")

    async def close(self) -> None:
        """Close the connection to Steam."""
        if self.is_closed():
            return

        self._closed = True

        if self.ws is not None:
            try:
                if self._state._active_auth_tickets:
                    self._state._active_auth_tickets.clear()
                    await self._state.deactivate_auth_session_tickets()
                await self.change_presence(apps=[])  # disconnect from games
                await self._state.handle_close()
            except ConnectionClosed:
                pass

        await self.http.close()
        self._ready.clear()

    def clear(self) -> None:
        """Clears the internal state of the bot. After this, the bot can be considered "re-opened", i.e.
        :meth:`is_closed` and :meth:`is_ready` both return ``False``. This also clears the internal cache.
        """
        self._closed = False
        self._ready.clear()
        self._state.clear()
        self.http.clear()
        if not self._aentered:
            self._tg = TaskGroup()

    async def _login(
        self,
        login_func: Callable[Concatenate[Self, P], Coroutine[None, None, SteamWebSocket]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        self.clear()
        async with nullcontext() if self._aentered else self._tg:
            state = self._state
            STATE.set(state)
            cm_list = None

            async def throttle() -> None:
                now = time.monotonic()
                between = now - last_connect
                sleep = random.random() * 4 if between > 300 else min(100 / between**0.5, 20)
                log.info("Attempting to connect to another CM in %ds", sleep)
                await asyncio.sleep(sleep)

            while not self.is_closed():
                last_connect = time.monotonic()

                try:
                    async with timeout(60):
                        self.ws = cast(
                            SteamWebSocket,
                            await login_func(
                                self,
                                *args,
                                **kwargs,
                                cm_list=cm_list,  # type: ignore
                            ),
                        )
                except RAISED_EXCEPTIONS:
                    if self.ws:
                        cm_list = self.ws.cm_list
                    await throttle()
                    continue

                try:
                    async with self.ws:
                        while True:
                            await state.ws.poll_event()

                except asyncio.CancelledError:  # KeyboardInterrupt
                    if not self.is_closed():
                        try:
                            await self.close()
                        except asyncio.CancelledError:
                            pass

                    return
                except RAISED_EXCEPTIONS:
                    self.dispatch("disconnect")
                    if not self.is_closed():
                        await throttle()

    @overload
    async def login(
        self,
        username: str,
        password: str,
        *,
        shared_secret: str = ...,
        identity_secret: str = ...,
    ) -> None: ...

    @overload
    async def login(
        self,
        *,
        refresh_token: str,
        shared_secret: str = ...,
        identity_secret: str = ...,
    ) -> None: ...

    async def login(
        self,
        username: str | None = None,
        password: str | None = None,
        *,
        shared_secret: str | None = None,
        identity_secret: str | None = None,
        refresh_token: str | None = None,
    ) -> None:
        """Initialize a connection to a Steam CM and login.

        If no ``shared_secret`` is passed, you will have to manually enter a Steam guard code using :meth:`code`.

        Parameters
        ----------
        username
            The username of the account to login to.
        password
            The password of the account to login to.
        shared_secret
            The shared secret for the account to login to.
        identity_secret
            The identity secret for the account to login to.
        refresh_token
            The refresh token of the account to login to.
        """
        self.username = username
        self.password = password
        self.shared_secret = shared_secret
        self.identity_secret = identity_secret

        if identity_secret is None:
            log.info("Trades will not be automatically accepted when sent as no identity_secret was passed.")

        await self._login(SteamWebSocket.from_client, refresh_token=refresh_token or self.refresh_token)

    async def anonymous_login(self) -> None:
        """Initialize a connection to a Steam CM and login anonymously."""
        await self._login(SteamWebSocket.anonymous_login_from_client)

    # state stuff
    def get_user(self, id: int, /) -> User | None:
        """Returns a user from cache with a matching ID or ``None`` if the user was not found.

        Parameters
        ----------
        id
            The ID64 of the user.
        """
        return self._state.get_user(_ID64_TO_ID32(id))

    async def fetch_user(self, id: int, /) -> User:
        """Fetches a user with a matching ID.

        Parameters
        ----------
        id
            The ID64 of the user.
        """
        return await self._state.fetch_user(ID64(id))

    async def fetch_users(self, *ids: int) -> Sequence[User]:
        """Fetches a list of :class:`~steam.User`.

        Note
        ----
        The :class:`~steam.User` objects returned are unlikely to retain the order they were originally in.

        Parameters
        ----------
        ids
            The user's ID64s.
        """
        return await self._state.fetch_users(cast("tuple[ID64, ...]", ids))

    def get_trade(self, id: int, /) -> TradeOffer | None:
        """Get a trade from cache with a matching ID or ``None`` if the trade was not found.

        Parameters
        ----------
        id
            The ID of the trade to search for from the cache.
        """
        return self._state.get_trade(TradeOfferID(id))

    async def fetch_trade(
        self, id: int, /, *, language: Language | None = None
    ) -> TradeOffer[Item[User], Item[ClientUser], User] | None:
        """Fetches a trade with a matching ID or ``None`` if the trade was not found.

        Parameters
        ----------
        id
            The ID of the trade to search for from the API.
        language
            The language to fetch the trade in. ``None`` uses the current language.
        """
        return await self._state.fetch_trade(TradeOfferID(id), language)

    def get_group(self, id: int, /) -> Group | None:
        """Get a group from cache with a matching ID or ``None`` if the group was not found.

        Parameters
        ----------
        id
            The ID64 of the group.
        """
        return self._state.get_group(_ID64_TO_ID32(id))

    def get_clan(self, id: int, /) -> Clan | None:
        """Get a clan from cache with a matching ID or ``None`` if the group was not found.

        Parameters
        ----------
        id
            The ID64 of the clan.
        """
        return self._state.get_clan(_ID64_TO_ID32(id))

    async def fetch_clan(self, id: int, /) -> Clan:
        """Fetches a clan from the websocket with a matching ID.

        Parameters
        ----------
        id
            The ID64 of the clan.
        """
        return await self._state.fetch_clan(ID64(id))

    async def create_clan(
        self,
        name: str,
        /,
        *,
        abbreviation: str | None = None,
        community_url: URL_ | str | None = None,
        public: bool = True,
        headline: str | None = None,
        summary: str | None = None,
        language: Language | None = None,
        country: str | None = None,
        state: str | None = None,
        city: str | None = None,
        apps: Iterable[App] | None = None,
        avatar: Media | None = None,
    ) -> Clan:
        """Create a clan.

        Parameters
        ----------
        name
            The name of the clan.
        avatar
            The avatar of the clan.
        abbreviation
            The abbreviation of the clan.
        community_url
            The community URL of the clan.
        public
            Whether the clan is public.
        headline
            The headline of the clan.
        summary
            The summary of the clan.
        language
            The language of the clan.
        country
            The country of the clan.
        state
            The state of the clan.
        city
            The city of the clan.
        apps
            The apps the clan is associated with.
        """
        if isinstance(community_url, str):
            community_url = URL_(community_url)
        if community_url is not None:
            community_url = community_url.parts[-1] or community_url.parts[-2]

        id64 = await self.http.create_clan(name, abbreviation, community_url, public)
        clan = await self._state._maybe_clan(id64, create=True)  # should just be created by steam for us
        await clan.edit(
            headline=headline,
            summary=summary,
            language=language,
            country=country,
            state=state,
            city=city,
            apps=apps,
            avatar=avatar,
        )
        try:
            await ChatGroup.join(clan)  # type: ignore  # should be removed one day
        except WSException:
            log.debug("Failed to join clan chat group", exc_info=True)
        return clan

    async def create_group(
        self,
        name: str,
        /,
        *,
        members: Iterable[IndividualID],
        avatar: Media | None = None,
        tagline: str | None = None,
    ) -> Group:
        """Create a group.

        Parameters
        ----------
        name
            The name of the group.
        members
            The members to add to the group.
        avatar
            The avatar of the group.
        tagline
            The tagline of the group.
        """
        group = await self._state.create_group(name, members)
        await group.edit(tagline=tagline, avatar=avatar)
        return group

    def get_app(self, id: int, /) -> PartialApp[None]:
        """Creates a :class:`PartialApp` instance from its ID.

        Parameters
        ----------
        id
            The app id of the app.
        """
        return PartialApp(self._state, id=id)

    async def fetch_app(self, id: int, /, *, language: Language | None = None) -> FetchedApp:
        """Fetch an app from its ID.

        Parameters
        ----------
        id
            The app id of the app.
        language
            The language to fetch the app in. If ``None`` uses the current language.

        Raises
        ------
        ValueError
            Passed `id` isn't for a valid app.
        """
        return await self._state.fetch_app(AppID(id), language)

    def get_package(self, id: int, /) -> PartialPackage:
        """Creates a :class:`PartialPackage` from its ID.

        Parameters
        ----------
        id
            The ID of the package.
        """
        return PartialPackage(self._state, id=id)

    async def fetch_package(self, id: int, /, *, language: Language | None = None) -> FetchedPackage:
        """Fetch a package from its ID.

        Parameters
        ----------
        id
            The ID of the package.
        language
            The language to fetch the package in. If ``None`` uses the current language.

        Raises
        ------
        ValueError
            Passed `id` isn't for a valid package.
        """
        return await self._state.fetch_package(PackageID(id), language)

    async def redeem_package(self, id: int, /) -> License:
        """Redeem a promotional free licenses package from its ID.

        Parameters
        ----------
        id
            The ID of the package to redeem.
        """
        id = PackageID(id)
        future: asyncio.Future[License] = asyncio.get_running_loop().create_future()
        async with self._state._license_lock:
            self._state.licenses_being_waited_for[id] = future
            await self.http.redeem_package(id)
            return await future

    def get_bundle(self, id: int, /) -> PartialBundle:
        """Creates a :class:`PartialBundle` instance from its ID.

        Parameters
        ----------
        id
            The ID of the bundle.
        """
        return PartialBundle(self._state, id=id)

    async def fetch_bundle(self, id: int, /, *, language: Language | None = None) -> FetchedBundle:
        """Fetch a bundle from its ID.

        Parameters
        ----------
        id
            The ID of the bundle.
        language
            The language to fetch the bundle in. If ``None`` uses the current language.
        """
        return await self._state.fetch_bundle(BundleID(id), language)

    @overload
    async def fetch_server(self, *, id: int) -> GameServer | None: ...

    @overload
    async def fetch_server(
        self,
        *,
        ip: IPAdress | str,
        port: int | None = None,
    ) -> GameServer | None: ...

    async def fetch_server(
        self,
        *,
        id: int | None = None,
        ip: IPAdress | str | None = None,
        port: int | str | None = None,
    ) -> GameServer | None:
        """Fetch a :class:`.GameServer` from its ip and port or its Steam ID or ``None`` if fetching the server failed.

        Parameters
        ----------
        ip
            The ip of the server.
        port
            The port of the server.
        id
            The ID64 of the game server. If passed, it makes a call to the master server to fetch its ip and port.

        Note
        ----
        Passing an ``ip``, ``port`` and ``id`` to this function will raise an :exc:`TypeError`.
        """

        if id is not None and ip is not None:
            raise TypeError("Too many arguments passed to fetch_server")
        if id:
            # we need to fetch the ip and port
            servers = await self._state.fetch_server_ip_from_steam_id(ID64(id))
            if not servers:
                raise ValueError(f"The master server didn't find a matching server for {id}")
            ip_, _, port = servers[0].addr.rpartition(":")
            ip = IPv4Address(ip_)
        elif not ip:
            raise TypeError("fetch_server missing argument ip")

        servers = await self.fetch_servers(Query.where(ip=f"{ip}{f':{port}' if port is not None else ''}"), limit=1)
        return servers[0] if servers else None

    async def fetch_servers(self, query: str | None = None, /, *, limit: int = 100) -> list[GameServer]:
        """Query game servers.

        Parameters
        ----------
        query
            The query to match servers with. If None returns all servers.
        limit
            The maximum amount of servers to return.

        See Also
        --------
        :meth:`.Query.where` to generate these programmatically.
        """
        servers = await self._state.fetch_servers(query or "", limit)
        return [GameServer(self._state, server) for server in servers]

    # content server related stuff

    @overload
    async def fetch_product_info(self, *, apps: Collection[App]) -> list[AppInfo]: ...

    @overload
    async def fetch_product_info(self, *, packages: Collection[Package]) -> list[PackageInfo]: ...

    @overload
    async def fetch_product_info(
        self, *, apps: Collection[App], packages: Collection[Package]
    ) -> tuple[list[AppInfo], list[PackageInfo]]: ...

    async def fetch_product_info(
        self, *, apps: Collection[App] = (), packages: Collection[Package] = ()
    ) -> list[AppInfo] | list[PackageInfo] | tuple[list[AppInfo], list[PackageInfo]]:
        """Fetch product info.

        Parameters
        ----------
        apps
            The apps to fetch info on.
        packages
            The packages to fetch info on.
        """

        app_infos, package_infos = await self._state.fetch_product_info(
            (app.id for app in apps), (package.id for package in packages)
        )

        if apps and packages:
            return app_infos, package_infos

        return app_infos if apps else package_infos

    @overload
    async def fetch_store_item(
        self, *, apps: Sequence[App], language: Language | None = None
    ) -> list[AppStoreItem]: ...

    @overload
    async def fetch_store_item(
        self, *, packages: Sequence[Package], language: Language | None = None
    ) -> list[PackageStoreItem]: ...

    @overload
    async def fetch_store_item(
        self, *, bundles: Sequence[Bundle], language: Language | None = None
    ) -> list[BundleStoreItem]: ...

    @overload
    async def fetch_store_item(
        self, *, apps: Sequence[App], packages: Sequence[Package], language: Language | None = None
    ) -> tuple[list[AppStoreItem], list[PackageStoreItem]]: ...

    @overload
    async def fetch_store_item(
        self, *, packages: Sequence[Package], bundles: Sequence[Bundle], language: Language | None = None
    ) -> tuple[list[PackageStoreItem], list[BundleStoreItem]]: ...

    @overload
    async def fetch_store_item(
        self, *, apps: Sequence[App], bundles: Sequence[Bundle], language: Language | None = None
    ) -> tuple[list[AppStoreItem], list[BundleStoreItem]]: ...

    @overload
    async def fetch_store_item(
        self,
        *,
        apps: Sequence[App],
        packages: Sequence[Package],
        bundles: Sequence[Bundle],
        language: Language | None = None,
    ) -> tuple[list[AppStoreItem], list[PackageStoreItem], list[BundleStoreItem]]: ...

    async def fetch_store_item(
        self,
        *,
        apps: Sequence[App] = (),
        packages: Sequence[Package] = (),
        bundles: Sequence[Bundle] = (),
        language: Language | None = None,
    ) -> Any:
        """Fetch store items.

        Parameters
        ----------
        apps
            The apps to fetch store items for.
        packages
            The packages to fetch store items for.
        bundles
            The bundles to fetch store items for.
        language
            The language to fetch the store items in. If ``None``, the default language is used.
        """
        if not any((apps, packages, bundles)):
            raise TypeError("fetch_store_info missing arguments apps, packages or bundles")
        resp = await self._state.fetch_store_info(
            (app.id for app in apps), (package.id for package in packages), (bundle.id for bundle in bundles), language
        )
        language = language or self.http.language
        match apps, packages, bundles:
            case [*_], (), ():  # simple cases
                return [AppStoreItem(self._state, proto, language) for proto in resp]
            case (), [*_], ():
                return [PackageStoreItem(self._state, proto, language) for proto in resp]
            case (), (), [*_]:
                return [BundleStoreItem(self._state, proto, language) for proto in resp]

            case [*_], [*_], ():  # mixed
                return [
                    AppStoreItem(self._state, proto, language)
                    for proto in resp
                    if proto.item_type == store.EStoreItemType.App
                ], [
                    PackageStoreItem(self._state, proto, language)
                    for proto in resp
                    if proto.item_type == store.EStoreItemType.Package
                ]
            case (), [*_], [*_]:
                return [
                    PackageStoreItem(self._state, proto, language)
                    for proto in resp
                    if proto.item_type == store.EStoreItemType.Package
                ], [
                    BundleStoreItem(self._state, proto, language)
                    for proto in resp
                    if proto.item_type == store.EStoreItemType.Bundle
                ]
            case [*_], (), [*_]:
                return [
                    AppStoreItem(self._state, proto, language)
                    for proto in resp
                    if proto.item_type == store.EStoreItemType.App
                ], [
                    BundleStoreItem(self._state, proto, language)
                    for proto in resp
                    if proto.item_type == store.EStoreItemType.Bundle
                ]
            case _:
                return (
                    [
                        AppStoreItem(self._state, proto, language)
                        for proto in resp
                        if proto.item_type == store.EStoreItemType.App
                    ],
                    [
                        PackageStoreItem(self._state, proto, language)
                        for proto in resp
                        if proto.item_type == store.EStoreItemType.Package
                    ],
                    [
                        BundleStoreItem(self._state, proto, language)
                        for proto in resp
                        if proto.item_type == store.EStoreItemType.Bundle
                    ],
                )

    async def register_cd_key(self, key: str, /) -> TransactionReceipt:
        """Register a CD key.

        Parameters
        ----------
        key
            The CD key to register.
        """
        proto = await self._state.register_cd_key(key)
        return TransactionReceipt(self._state, proto)

    # miscellaneous stuff

    async def fetch_published_file(
        self,
        id: int,
        /,
        *,
        revision: PublishedFileRevision = PublishedFileRevision.Default,
        language: Language | None = None,
    ) -> PublishedFile | None:
        """Fetch a published file from its ID.

        Parameters
        ----------
        id
            The ID of the published file.
        revision
            The revision of the published file to fetch.
        language
            The language to fetch the published file in. If ``None``, the current language is used.
        """
        (file,) = await self._state.fetch_published_files((PublishedFileID(id),), revision, language)
        return file

    async def fetch_published_files(
        self,
        *ids: int,
        revision: PublishedFileRevision = PublishedFileRevision.Default,
        language: Language | None = None,
    ) -> list[PublishedFile | None]:
        """Fetch published files from their IDs.

        Parameters
        ----------
        ids
            The IDs of the published files.
        revision
            The revision of the published files to fetch.
        language
            The language to fetch the published files in. If ``None``, the current language is used.
        """
        return await self._state.fetch_published_files(cast(tuple[PublishedFileID, ...], ids), revision, language)

    async def create_post(self, content: str, /, app: App | None = None) -> Post[ClientUser]:
        """Create a post.

        Parameters
        ----------
        content
            The content of the post.
        app
            The app to create the post for.
        """
        await self._state.create_user_post(content, app_id=AppID(0) if app is None else app.id)
        async for entry in self.user_news(flags=UserNewsType.Friend, app=app):
            if entry.app == app and entry.actor == self.user:
                post = await entry.post()
                if post.content == content:
                    return cast("Post[ClientUser]", post)
        raise RuntimeError("Post created has no ID, this should be unreachable")

    async def user_news(
        self,
        *,
        limit: int | None = None,
        before: datetime.datetime | None = None,
        after: datetime.datetime | None = None,
        app: App | None = None,
        flags: UserNewsType | None = None,
        language: Language | None = None,
    ) -> AsyncGenerator[UserNews, None]:
        """Fetch news for the user.

        Parameters
        ----------
        limit
            The maximum number of news entries to fetch.
        before
            The date to fetch news before.
        after
            The date to fetch news after.
        app
            The app to fetch news entries related to.
        flags
            The type of news to fetch.
        language
            The language to fetch the news in. If ``None``, the current language is used.
        """
        before = original_before = before or DateTime.now()
        after = after or UNIX_EPOCH
        if flags is None:
            flags = UserNewsType.Friend if app is None else UserNewsType.App

        while True:
            msg = await self._state.fetch_user_news(flags, app.id if app is not None else None, before, after, language)

            achievements = {
                achievements.appid: {
                    achievement.name: UserNewsAchievement(
                        achievement.name,
                        PartialApp(self._state, id=achievements.appid),
                        achievement.display_name,
                        achievement.display_description,
                        CDNAsset(
                            self._state,
                            f"{URL.CDN}/steamcommunity/public/images/apps/{achievements.appid}/{achievement.icon}",
                        ),
                        achievement.hidden,
                        round(achievement.unlocked_pct, 1),
                    )
                    for achievement in achievements.achievements
                }
                for achievements in msg.achievement_display_data
            }
            if not msg.news:
                return

            for news_item in msg.news:
                news = UserNews(self._state, news_item, achievements.get(news_item.gameid, {}))
                if after < news.created_at < original_before:
                    yield news
                else:
                    return
                before = news.created_at
                if limit is not None:
                    limit -= 1
                    if limit <= 0:
                        return

    async def trade_history(
        self,
        *,
        limit: int | None = 100,
        before: datetime.datetime | None = None,
        after: datetime.datetime | None = None,
        language: Language | None = None,
        include_failed: bool = True,
    ) -> AsyncGenerator[TradeOffer[MovedItem[User], MovedItem[ClientUser], User], None]:
        """An :term:`asynchronous iterator` for accessing a :class:`steam.ClientUser`'s
        :class:`steam.TradeOffer` objects.

        Examples
        --------

        Usage:

        .. code:: python

            async for trade in client.trade_history(limit=10):
                items = [getattr(item, "name", str(item.id)) for item in trade.receiving]
                items = ", ".join(items) or "Nothing"
                print("Partner:", trade.user)
                print("Sent:", items)

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of trades to search through. Default is ``100``. Setting this to ``None`` will fetch all
            of the user's trades, but this will be a very slow operation.
        before
            A time to search for trades before.
        after
            A time to search for trades after.
        language
            The language to fetch the trade in. ``None`` uses the current language.
        include_failed
            Whether to include trades that failed.

        Yields
        ---------
        :class:`~steam.TradeOffer`
        """
        from .trade import TradeOffer

        total: int | None = None
        previous_time = 0
        after = after or UNIX_EPOCH
        before = before or DateTime.now()
        after_timestamp = after.timestamp()
        before_timestamp = before.timestamp()
        yielded = 0

        async def get_trades(
            page: int = 100,
        ) -> list[TradeOffer[MovedItem[User], MovedItem[ClientUser], User]]:
            nonlocal total, previous_time
            data = await self.http.get_trade_history(page, include_failed, previous_time, language)
            if total is None:
                total = data.get("total_trades", 0)
            if not total:
                return []

            descriptions = data.get("descriptions", ())
            trades = [
                TradeOffer._from_history(self._state, trade, descriptions)
                for trade in data.get("trades", ())
                if after_timestamp < trade["time_init"] < before_timestamp
            ]
            previous_created_at = trades[-1].created_at
            assert previous_created_at is not None
            previous_time = previous_created_at.timestamp()
            for trade, partner in zip(trades, await self._state._maybe_users(trade.user.id64 for trade in trades)):
                trade.user = partner
                for item in trade.receiving:
                    item.owner = partner
            return cast(
                "list[TradeOffer[MovedItem[User], MovedItem[ClientUser], User]]", trades
            )  # this should be safe at this point

        for trade in await get_trades():
            if limit is not None and yielded >= limit:
                return
            yield trade
            yielded += 1

        assert total is not None
        if total < 100:
            for page in range(200, math.ceil((total + 100) / 100) * 100, 100):
                for trade in await get_trades(page):
                    if limit is not None and yielded >= limit:
                        return
                    yield trade
                    yielded += 1

    async def all_apps(
        self,
        *,
        limit: int | None = None,
        modified_after: datetime.datetime | None = None,
        include_games: bool = True,
        include_dlc: bool = True,
        include_software: bool = True,
        include_videos: bool = True,
        include_hardware: bool = True,
    ) -> AsyncGenerator[AppListApp, None]:
        """An :term:`asynchronous iterator` over all the apps on Steam.

        Parameters
        ----------
        limit
            The maximum number of apps to search through. Default is ``None``. Setting this to ``None`` will fetch all
            of the apps, but this will be a very slow operation.
        modified_after
            A time to search for apps after.
        include_games
            Whether to include games.
        include_dlc
            Whether to include DLC.
        include_software
            Whether to include software.
        include_videos
            Whether to include videos.
        include_hardware
            Whether to include hardware.

        Yields
        ------
        :class:`~steam.AppListApp`
        """
        async for app in self.http.get_all_apps(
            include_games,
            include_dlc,
            include_software,
            include_videos,
            include_hardware,
            limit=limit,
            chunk_size=min(limit if limit is not None else 10_000, 50_000),
            modified_after=modified_after,
        ):
            yield AppListApp(self._state, app)

    @utils.todo
    async def all_tags(self) -> Any:
        raise NotImplementedError

    @utils.todo
    async def all_categories(self) -> Any:
        raise NotImplementedError

    @utils.todo
    async def all_genres(self) -> Any:
        raise NotImplementedError

    @utils.todo
    async def all_stickers(self) -> Any:
        raise NotImplementedError  # QueryRewardItemsRequest

    async def change_presence(
        self,
        *,
        app: App | None = None,
        apps: Iterable[App] | None = None,
        state: PersonaState | None = None,
        ui_mode: UIMode | None = None,
        flags: PersonaStateFlag | None = None,
        force_kick: bool = False,
    ) -> None:
        """Set your status.

        Parameters
        ----------
        app
            An app to set your status as.
        apps
            A list of apps to set your status to.
        state
            The state to show your account as.

            Warning
            -------
            Setting your status to :attr:`~steam.PersonaState.Offline`, will stop you receiving persona state updates
            and by extension :meth:`on_user_update` will stop being dispatched.

        ui_mode
            The UI mode to set your status to.
        flags
            The flags to update your account with.
        force_kick
            Whether to forcefully kick any other playing sessions.
        """
        apps_ = [app.to_proto() for app in apps] if apps is not None else []
        if app is not None:
            apps_.append(app.to_proto())
        if self.ws is None:
            raise RuntimeError("Client is not logged in")
        await self.ws.change_presence(apps=apps_, state=state, flags=flags, ui_mode=ui_mode, force_kick=force_kick)

    async def trade_url(self, generate_new: bool = False) -> TradeURLInfo:
        """Fetches this account's trade url.

        Parameters
        ----------
        generate_new
            Whether or not to generate a new trade token, defaults to ``False``.
        """
        info = utils.parse_trade_url(await self._state.fetch_trade_url(generate_new))
        assert info is not None
        return info

    async def wait_until_ready(self) -> None:
        """Waits until the client's internal cache is all ready."""
        await self._ready.wait()

    async def fetch_price(self, name: str, app: App, currency: Currency | None = None) -> PriceOverview:
        """Fetch the :class:`PriceOverview` for an item.

        Parameters
        ----------
        name
            The name of the item.
        app
            The app the item is from.
        currency
            The currency to fetch the price in.
        """
        price = await self.http.get_price(app.id, name, currency)
        return PriceOverview(price, currency or Currency.USD)

    # events to be subclassed

    async def on_error(self, event: str, error: Exception, *args: object, **kwargs: object):
        """The default error handler provided by the client.

        Usually when an event raises an uncaught exception, a traceback is printed to :attr:`sys.stderr` and the
        exception is ignored. If you want to change this behaviour and handle the exception yourself, this event can
        be overridden. Which, when done, will suppress the default action of printing the traceback.

        If you want exception to propagate out of the :class:`Client` class you can define an ``on_error`` handler
        consisting of a single empty :ref:`py:raise`. Exceptions raised by ``on_error`` will not be handled in any
        way by :class:`Client`.

        Parameters
        ----------
        event
            The name of the event that errored.
        error
            The error that was raised.
        args
            The positional arguments associated with the event.
        kwargs
            The key-word arguments associated with the event.
        """
        print(f"Ignoring exception in {event}", file=sys.stderr)
        traceback.print_exception(error, file=sys.stderr)

    if TYPE_CHECKING or DOCS_BUILDING:
        # these methods shouldn't exist at runtime unless subclassed to prevent pollution of logs

        async def on_connect(self) -> None:
            """Called when the client has successfully connected to Steam. This is not the same as the client being
            fully prepared, see :func:`on_ready` for that.

            The warnings on :meth:`on_ready` also apply.
            """

        async def on_disconnect(self) -> None:
            """Called when the client has disconnected from Steam. This could happen either through the internet
            disconnecting, an explicit call to logout, or Steam terminating the connection.

            This function can be called multiple times.
            """

        async def on_ready(self) -> None:
            """Called after a successful login and the client has handled setting up everything.

            Warning
            -------
            This function is not guaranteed to be the first event called. Likewise, this function is **not** guaranteed
            to only be called once. This library implements reconnection logic and will therefore end up calling this
            event whenever a CM disconnects.
            """

        async def on_login(self) -> None:
            """Called when the client has logged into https://steamcommunity.com."""

        async def on_logout(self) -> None:
            """Called when the client has logged out of https://steamcommunity.com."""

        async def on_message(self, message: steam.Message, /) -> None:
            """Called when a message is created.

            Parameters
            ----------
            message
                The message that was received.
            """

        async def on_typing(self, user: steam.User, when: datetime.datetime, /) -> None:
            """Called when typing is started.

            Parameters
            ----------
            user
                The user that started typing.
            when
                The time the user started typing at.
            """

        async def on_reaction_add(self, reaction: MessageReaction, /) -> None:
            """Called when a reaction is added to a message.

            Parameters
            ----------
            reaction
                The reaction that was added.
            """

        async def on_reaction_remove(self, reaction: MessageReaction, /) -> None:
            """Called when a reaction is removed from a message.

            Parameters
            ----------
            reaction
                The reaction that was removed.
            """

        async def on_trade(self, trade: steam.TradeOffer, /) -> None:
            """Called when the client sends/receives a trade offer.

            Parameters
            ----------
            trade
                The trade offer that was received/send.
            """

        async def on_trade_update(self, before: steam.TradeOffer, after: steam.TradeOffer, /) -> None:
            """Called when the client or the trade partner updates a trade offer.

            Parameters
            ----------
            before
                The trade offer that was updated prior to the update.
            after
                The trade offer now.
            """

        async def on_comment(self, comment: steam.Comment, /) -> None:
            """Called when the client receives a comment notification.

            Parameters
            ----------
            comment
                The comment received.
            """

        async def on_invite(self, invite: steam.Invite, /) -> None:
            """Called when the client receives/sends an invitation.

            Parameters
            ----------
            invite
                The invite received.
            """

        async def on_invite_accept(self, invite: steam.Invite, /) -> None:
            """Called when the client/author accepts an invitation.

            Parameters
            ----------
            invite
                The invite that was accepted.
            """

        async def on_invite_decline(self, invite: steam.Invite, /) -> None:
            """Called when the client/author declines an invitation.

            Parameters
            ----------
            invite
                The invite that was declined.
            """

        async def on_user_update(self, before: steam.User, after: steam.User, /) -> None:
            """Called when a user is updated, due to one or more of the following attributes changing:

                - :attr:`~steam.User.name`
                - :attr:`~steam.User.state`
                - :attr:`~steam.User.flags`
                - :attr:`~steam.User.avatar_url`
                - :attr:`~steam.User.last_logon`
                - :attr:`~steam.User.last_logoff`
                - :attr:`~steam.User.last_seen_online`
                - :attr:`~steam.User.app`

            Parameters
            ----------
            before
                The user's state before it was updated.
            after
                The user's state now.
            """

        async def on_friend_add(self, friend: steam.Friend, /) -> None:
            """Called when a friend is added to the client's friends list.

            Parameters
            ----------
            friend
                The friend that was added.
            """

        async def on_friend_remove(self, friend: steam.Friend, /) -> None:
            """Called when you or the ``friend`` remove each other from your friends lists.

            Parameters
            ----------
            friend
                The friend who was removed.
            """

        async def on_clan_join(self, clan: steam.Clan, /) -> None:
            """Called when the client joins a new clan.

            Parameters
            ----------
            clan
                The joined clan.
            """

        async def on_clan_update(self, before: steam.Clan, after: steam.Clan, /) -> None:
            """Called when a clan is updated, due to one or more of the following attributes changing:

                - :attr:`~steam.Clan.name`
                - :attr:`~steam.Clan.avatar_url`
                - :attr:`~steam.Clan.member_count`
                - :attr:`~steam.Clan.in_game_count`
                - :attr:`~steam.Clan.online_count`
                - :attr:`~steam.Clan.active_member_count`

            Parameters
            ----------
            before
                The clan's state before it was updated.
            after
                The clan's state now.
            """

        async def on_clan_leave(self, clan: steam.Clan, /) -> None:
            """Called when the client leaves a clan.

            Parameters
            ----------
            clan
                The left clan.
            """

        async def on_group_join(self, group: steam.Group, /) -> None:
            """Called when the client joins a new group.

            Parameters
            ----------
            group
                The joined group.
            """

        async def on_group_update(self, before: steam.Group, after: steam.Group, /) -> None:
            """Called when a group is updated.

            Parameters
            ----------
            before
                The group's state before it was updated.
            after
                The group's state now.
            """

        async def on_group_leave(self, group: steam.Group, /) -> None:
            """Called when the client leaves a group.

            Parameters
            ----------
            group
                The left group.
            """

        async def on_event_create(self, event: steam.Event, /) -> None:
            """Called when an event in a clan is created.

            Parameters
            ----------
            event
                The event that was created.
            """

        async def on_announcement_create(self, announcement: steam.Announcement, /) -> None:
            """Called when an announcement in a clan is created.

            Parameters
            ----------
            announcement
                The announcement that was created.
            """

        async def on_authentication_ticket_update(
            self, ticket: steam.AuthenticationTicket, response: steam.AuthSessionResponse, state: int, /
        ) -> None:
            """Called when the client's authentication ticket is updated.

            Parameters
            ----------
            ticket
                The updated ticket.
            response
                The response code from the CM.
            state
                The state of the ticket.
            """

    @overload
    async def wait_for(
        self,
        event: Literal[
            "connect",
            "disconnect",
            "ready",
            "login",
            "logout",
        ],
        *,
        check: Callable[[], bool] = ...,
        timeout: float | None = ...,
    ) -> None: ...

    @overload
    async def wait_for(
        self,
        event: Literal["error"],
        *,
        check: Callable[[str, Exception, tuple[Any, ...], dict[str, Any]], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[str, Exception, tuple[Any, ...], dict[str, Any]]: ...

    @overload
    async def wait_for(
        self,
        event: Literal["message"],
        *,
        check: Callable[[Message], bool] = ...,
        timeout: float | None = ...,
    ) -> Message: ...

    @overload
    async def wait_for(
        self,
        event: Literal["comment"],
        *,
        check: Callable[[Comment], bool] = ...,
        timeout: float | None = ...,
    ) -> Comment: ...

    @overload
    async def wait_for(
        self,
        event: Literal["user_update"],
        *,
        check: Callable[[User, User], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[User, User]: ...

    @overload
    async def wait_for(
        self,
        event: Literal["clan_update"],
        *,
        check: Callable[[Clan, Clan], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[Clan, Clan]: ...

    @overload
    async def wait_for(
        self,
        event: Literal["group_update"],
        *,
        check: Callable[[Group, Group], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[Group, Group]: ...

    @overload
    async def wait_for(
        self,
        event: Literal["typing"],
        *,
        check: Callable[[User, datetime.datetime], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[User, datetime.datetime]: ...

    @overload
    async def wait_for(
        self,
        event: Literal["trade"],
        *,
        check: Callable[[TradeOffer], bool] = ...,
        timeout: float | None = ...,
    ) -> TradeOffer: ...

    @overload
    async def wait_for(
        self,
        event: Literal["trade_update",],
        *,
        check: Callable[[TradeOffer, TradeOffer], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[TradeOffer, TradeOffer]: ...

    @overload
    async def wait_for(
        self,
        event: Literal[
            "friend_add",
            "friend_remove",
        ],
        *,
        check: Callable[[Friend], bool] = ...,
        timeout: float | None = ...,
    ) -> User: ...

    @overload
    async def wait_for(
        self,
        event: Literal[
            "invite",
            "invite_accept",
            "invite_decline",
        ],
        *,
        check: Callable[[UserInvite | ClanInvite | GroupInvite | AppInvite], bool] = ...,
        timeout: float | None = ...,
    ) -> UserInvite | ClanInvite | GroupInvite | AppInvite: ...

    @overload
    async def wait_for(
        self,
        event: Literal[
            "clan_join",
            "clan_leave",
        ],
        *,
        check: Callable[[Clan], bool] = ...,
        timeout: float | None = ...,
    ) -> Clan: ...

    @overload
    async def wait_for(
        self,
        event: Literal[
            "group_join",
            "group_leave",
        ],
        *,
        check: Callable[[Group], bool] = ...,
        timeout: float | None = ...,
    ) -> Group: ...

    @overload
    async def wait_for(
        self,
        event: Literal["event_create"],
        *,
        check: Callable[[Event], bool] = ...,
        timeout: float | None = ...,
    ) -> Event: ...

    @overload
    async def wait_for(
        self,
        event: Literal["announcement_create"],
        *,
        check: Callable[[Announcement], bool] = ...,
        timeout: float | None = ...,
    ) -> Announcement: ...

    @overload
    async def wait_for(
        self,
        event: Literal["authentication_ticket_update"],
        *,
        check: Callable[[AuthenticationTicket, AuthSessionResponse, int], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[AuthenticationTicket, AuthSessionResponse, int]: ...

    @overload
    async def wait_for(
        self: Bot,
        event: Literal["command_error"],
        *,
        check: Callable[[commands.Context, Exception], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[commands.Context, Exception]: ...

    @overload
    async def wait_for(
        self: Bot,
        event: Literal[
            "command",
            "command_completion",
        ],
        *,
        check: Callable[[commands.Context], bool] = ...,
        timeout: float | None = ...,
    ) -> commands.Context: ...

    async def wait_for(
        self,
        event: str,
        *,
        check: Callable[..., bool] = return_true,
        timeout: float | None = None,
    ) -> Any:
        """Wait for the first event to be dispatched that meets the requirements, this by default is the first event
        with a matching event name.

        Parameters
        ----------
        event
            The event name from the :ref:`event reference <event-reference>`, but without the ``on_`` prefix, to wait
            for.
        check
            A callable predicate that checks the received event. The arguments must match the parameters of the
            ``event`` being waited for and must return a :class:`bool`.
        timeout
            By default, :meth:`wait_for` function does not timeout, however, in the case a ``timeout`` parameter is
            passed after the amount of seconds passes :exc:`asyncio.TimeoutError` is raised.

        Raises
        ------
        :exc:`asyncio.TimeoutError`
            If the provided timeout was reached.

        Returns
        -------
        Returns ``None``, a single argument or a :class:`tuple` of multiple arguments that mirrors the parameters for
        the ``event`` parameter from the :ref:`event reference <event-reference>`.
        """
        future = asyncio.get_running_loop().create_future()

        event_lower = event.lower()
        try:
            listeners = self._listeners[event_lower]
        except KeyError:
            listeners = []
            self._listeners[event_lower] = listeners

        listeners.append((future, check))
        return await asyncio.wait_for(future, timeout)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, TypedDict

if TYPE_CHECKING:
    from typing_extensions import NotRequired


class UserRecentlyPlayedApp(TypedDict):
    name: str
    appid: int
    playtime_2weeks: int
    playtime_forever: int
    img_icon_url: str


class WishlistApp(TypedDict):
    name: str
    capsule: str
    review_score: int
    review_desc: str
    reviews_total: str
    reviews_percent: int
    release_date: str | None  # really an int/float (sometimes)
    release_string: str
    platform_icons: str
    subs: list[WishlistAppSub]
    type: str
    screenshots: list[str]
    review_css: str
    priority: int
    added: int
    background: str
    rank: int
    tags: list[str]
    is_free_game: bool
    win: bool
    mac: bool
    linux: bool


class WishlistAppSub(TypedDict):
    packageid: int
    bundleid: int
    discount_block: str
    discount_pct: int
    price: int


class PackageGroups(TypedDict):
    name: Literal["default", "subscriptions"]
    title: str
    description: str
    selection_text: str
    display_type: Literal[0, 1]
    is_recurring_subscription: bool
    subs: list[PackageGroupSub]


class PackageGroupSub(TypedDict):
    packageid: int
    percent_savings_text: str
    percent_savings: int
    option_text: str
    option_description: str
    can_get_free_license: str
    is_free_license: bool
    price_in_cents_with_discount: int


class FetchedAppPriceOverview(TypedDict):
    currency: str
    initial: int
    final: int
    discount_percent: int
    initial_formatted: str
    final_formatted: str


class FetchedAppCategory(TypedDict):
    id: int
    description: str


class FetchedApp(TypedDict):
    # https://wiki.teamfortress.com/wiki/User:RJackson/StorefrontAPI#Result_data_3
    type: Literal["game", "dlc", "demo", "advertising", "mod", "video"]
    name: str
    steam_appid: int
    required_age: int
    is_free: bool
    controller_support: Literal["partial", "full"]
    dlc: list[int]
    detailed_description: str
    short_description: str
    fullgame: NotRequired[dict[str, Any]]
    supported_languages: str
    header_image: str
    pc_requirements: list[dict[str, str]]
    mac_requirements: list[dict[str, str]]
    linux_requirements: list[dict[str, str]]
    legal_notice: str
    developers: list[str]
    publishers: list[str]
    demos: list[dict[str, Any]]
    price_overview: FetchedAppPriceOverview
    packages: list[int]
    package_groups: list[PackageGroups]
    platforms: dict[str, bool]
    metacritic: list[dict[str, str]]
    categories: list[FetchedAppCategory]
    release_date: dict[str, str]
    background: str
    website: str
    movies: list[dict[str, Any]]
    content_descriptors: dict[str, list[int]]  # a lie but who cares the "notes" key isn't useful


class DLCPriceOverview(TypedDict):
    currency: str
    initial: int
    final: int
    discount_percent: int


class DLCPlatforms(TypedDict):
    windows: bool
    mac: bool
    linux: bool


class DLCReleaseDate(TypedDict):
    steam: str  # unix timestamps
    mac: str
    linux: str


class DLC(TypedDict):
    id: int
    name: str
    header_image: str
    price_overview: DLCPriceOverview
    platforms: DLCPlatforms
    release_date: DLCReleaseDate


class Leaderboard(TypedDict):
    id: int
    name: str
    display_name: str
    entry_count: int
    sort_method: int
    display_type: int


class GetAppList(TypedDict):
    apps: list[AppListApp]
    have_more_results: bool
    last_appid: int


class AppListApp(TypedDict):
    appid: int
    name: str
    last_modified: int
    price_change_number: int


class AssetPricesAsset(TypedDict("", {"class": list[dict[str, Any]]})):
    prices: dict[str, int]  # CurrencyCode.name => price in base denomination (£1.54 == 154)
    original_prices: NotRequired[dict[str, int]]
    name: str  # really the def_index?
    date: str  # the time the item's price was last changed?
    classid: str
    tags: NotRequired[list[str]]
    tag_ids: NotRequired[list[int]]


class AssetPrices(TypedDict):
    success: bool
    assets: list[AssetPricesAsset]
    tags: NotRequired[dict[str, str]]
    tag_ids: NotRequired[dict[int, int]]  # only use .values()



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import bisect
import math
from collections import Counter
from collections.abc import Mapping
from dataclasses import dataclass
from datetime import timedelta
from operator import itemgetter
from typing import TYPE_CHECKING, Final, Generic, TypeVar, cast, overload

from ... import abc, user, utils
from ..._const import timeout
from ..._gc.client import ClientUser as ClientUser_
from ...id import parse_id64
from ...types.id import ID32
from ...utils import DateTime
from ..commands.converters import Converter, UserConverter
from .enums import Rank, RankType
from .protobufs import cstrike

if TYPE_CHECKING:
    from typing import Literal

    from typing_extensions import Self

    from ...app import CSGO, App
    from ...enums import Language
    from ...trade import Inventory, Item
    from .backpack import Backpack
    from .state import GCState


UserT = TypeVar("UserT", bound=abc.PartialUser)

__all__ = (
    "Team",
    "Round",
    "Match",
    "PartialUser",
    "User",
    "ClientUser",
    "MatchPlayer",
    "ProfileInfo",
)


class MatchWatchInfo: ...


@dataclass(slots=True)
class Team:
    """Represents a team in a match"""

    score: int
    won: bool
    players: list[MatchPlayer]


@dataclass(slots=True)
class Round:
    """Represents a round of a match"""

    duration: timedelta
    teams: list[Team]
    map: str

    @property
    def players(self) -> list[MatchPlayer]:
        return [player for team in self.teams for player in team.players]


class Match:
    """Represents a match of CSGO"""

    def __init__(self, state: GCState, match_info: cstrike.MatchInfo, players: dict[ID32, User]) -> None:
        self._state = state
        self.id = match_info.matchid
        self.created_at = DateTime.from_timestamp(match_info.matchtime)
        # self.watch_info = MatchWatchInfo(match_info.watchablematchinfo)
        # self.type = self.watch_info.type
        # self.map_group = self.watch_info.map_group
        # self.map = self.watch_info.map
        # self.server_id = self.watch_info.server_id

        self.rounds: list[Round] = []
        for round in match_info.roundstatsall:
            player_ids = cast(list[ID32], round.reservation.account_ids)
            team_size = len(player_ids) // len(round.team_scores)
            previous_scores = [0] * len(round.team_scores)
            teams: list[Team] = []
            for idx, score in enumerate(round.team_scores):
                players_: list[MatchPlayer] = []
                for id in player_ids[(idx * team_size) : (idx + 1) * team_size]:
                    player = MatchPlayer(state, players[id])
                    idx = player_ids.index(id)
                    player.kills = round.kills[idx]
                    player.assists = round.assists[idx]
                    player.deaths = round.deaths[idx]
                    player.score = round.scores[idx]
                    player.enemy_kills = round.enemy_kills[idx]
                    player.enemy_head_shots = round.enemy_headshots[idx]
                    players_.append(player)

                won = round.team_scores[idx] > previous_scores[idx]
                if won:
                    mvp_idx, _ = max(
                        Counter(round.mvps[(idx * team_size) : (idx + 1) * team_size]).items(), key=itemgetter(1)
                    )  # can't have an even number of players right?
                    mvp = utils.get(players_, id=player_ids[mvp_idx])
                    assert mvp
                    mvp.mvp = True
                teams.append(Team(score, won, players_))
            self.rounds.append(Round(timedelta(seconds=round.match_duration), teams, round.map))

    @property
    def players(self) -> list[MatchPlayer]:
        """All the players that appeared in the match."""
        return list(dict.fromkeys(player for round in reversed(self.rounds) for player in round.players))

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id}>"


@dataclass
class Matches:
    matches: list[Match]
    streams: list[cstrike.TournamentTeam]
    tournament_info: cstrike.TournamentInfo


class PartialUser(abc.PartialUser):
    __slots__ = ()
    _state: GCState

    async def csgo_profile(self) -> ProfileInfo[Self]:
        """Fetches this users CSGO profile info."""
        msg = await self._state.fetch_user_csgo_profile(self.id)
        if not msg.account_profiles:
            raise ValueError
        return ProfileInfo(self, msg.account_profiles[0])

    async def recent_matches(self) -> Matches:
        """Fetches this user's recent games."""
        future = self._state.ws.gc_wait_for(
            cstrike.MatchList,
            check=lambda msg: (
                msg.msgrequestid == cstrike.MatchListRequestRecentUserGames.MSG and msg.accountid == self.id
            ),
        )
        await self._state.ws.send_gc_message(cstrike.MatchListRequestRecentUserGames(accountid=self.id))
        async with timeout(30):
            msg = await future

        players = {
            user.id: user
            for user in await self._state._maybe_users(
                parse_id64(id)
                for match in msg.matches
                for round in match.roundstatsall
                for id in round.reservation.account_ids
            )
        }

        return Matches([Match(self._state, match, players) for match in msg.matches], msg.streams, msg.tournamentinfo)


class User(PartialUser, user.User):  # type: ignore
    __slots__ = ()


class ClientUser(PartialUser, ClientUser_):  # type: ignore
    __slots__ = ("_profile_info_msg",)

    if TYPE_CHECKING:

        @overload
        async def inventory(self, app: Literal[CSGO], *, language: object = ...) -> Backpack:  # type: ignore
            ...

        @overload
        async def inventory(self, app: App, *, language: Language | None = None) -> Inventory[Item[Self], Self]:  # type: ignore
            ...

    async def csgo_profile(self) -> ProfileInfo[Self]:
        return ProfileInfo(self, self._profile_info_msg)

    @utils.todo
    async def live_games(self) -> ...: ...


class MatchPlayer(PartialUser, user.WrapsUser):  # type: ignore
    kills: int
    assists: int
    deaths: int
    score: int
    enemy_kills: int
    enemy_head_shots: int
    mvp: bool


LEVEL_MAP: Final = cast(Mapping[int, str], {
    0: "Not Recruited",
    1: "Recruit",  # 1-3
    4: "Private",  # 4-7 etc.
    8: "Corporal",
    12: "Sergeant",
    16: "Master Sergeant",
    20: "Sergeant Major",
    24: "Lieutenant",
    28: "Captain",
    32: "Major",
    35: "Colonel",
    36: "Brigadier General",
    37: "Major General",
    38: "Lieutenant General",
    39: "General",
    40: "Global General",
})  # fmt: skip


class Level(int):
    @property
    def name(self, levels: list[int] = list(LEVEL_MAP), names: list[str] = list(LEVEL_MAP.values()), /) -> str:
        return names[bisect.bisect_left(levels, self)]


class Ranking:
    __slots__ = ("rank", "id", "rank_type", "change", "wins", "tv_control")

    def __init__(self, proto: cstrike.PlayerRankingInfo) -> None:
        self.rank = Rank.try_value(proto.rank_id) if proto.rank_type_id != 10 else proto.rank_id
        self.id = proto.rank_type_id
        self.rank_type = RankType.try_value(proto.rank_type_id)
        self.change = proto.rank_change
        self.wins = proto.wins
        self.tv_control = proto.tv_control


class ProfileInfo(Generic[UserT]):
    def __init__(self, user: UserT, proto: cstrike.MatchmakingClientHello | cstrike.PlayersProfileProfile):
        self.user = user
        self.in_match = proto.ongoingmatch
        self.global_stats = proto.global_stats
        self.penalty_seconds = proto.penalty_seconds
        self.penalty_reason = proto.penalty_reason
        self.vac_banned = proto.vac_banned
        self.ranking = Ranking(proto.ranking)
        self.commendation = proto.commendation
        self.medals = proto.medals
        self.current_event = proto.my_current_event
        self.current_event_teams = proto.my_current_event_teams
        self.current_team = proto.my_current_team
        self.current_event_stages = proto.my_current_event_stages
        self.survey_vote = proto.survey_vote
        self.activity = proto.activity
        self.current_xp = proto.player_cur_xp
        self.level = Level(proto.player_level)
        self.xp_bonus_flags = proto.player_xp_bonus_flags
        self.rankings = [Ranking(ranking) for ranking in proto.rankings]

    @property
    def percentage_of_current_level(self) -> int:
        """The user's current percentage of their current level."""
        return math.floor(max(self.current_xp - 327680000, 0) / 5000)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} user={self.user!r}>"


class CSGOUserConverter(Converter[User]):
    convert = UserConverter.convert  # type: ignore



"""
steam.protobufs
~~~~~~~~~~~~~~~
The protobufs for the Steam API.

Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE
"""

from collections.abc import Mapping
from typing import TYPE_CHECKING, Final, TypeAlias, cast

import betterproto

PROTOBUFS: Final = cast("Mapping[EMsg, type[ProtobufMessage | Message]]", {})
RequestType: TypeAlias = "type[UnifiedMessage]"
ResponseType: TypeAlias = "type[UnifiedMessage]"
UMS: Final = cast(Mapping[str, tuple[RequestType, ResponseType]], {})
GC_PROTOBUFS: Final = cast("Mapping[AppID, Mapping[IntEnum, type[GCProtobufMessage | GCMessage]]]", {})

from .emsg import *
from .msg import *

if TYPE_CHECKING:
    from ..enums import IntEnum
    from ..types.id import AppID

__all__ = (
    "EMsg",
    "NoMsg",
    "Message",
    "ProtobufMessage",
    "UnifiedMessage",
    "GCProtobufMessage",
    "GCMessage",
    "PROTOBUFS",
    "UMS",
    "GC_PROTOBUFS",
    "REQUEST_EMSGS",
    "RESPONSE_EMSGS",
    "SERVICE_EMSGS",
)

# NOTE all modules need to be included here otherwise messages won't be parsed correctly
from . import (
    app_info as app_info,
    auth as auth,
    base as base,
    chat as chat,
    clan as clan,
    client_server as client_server,
    client_server_2 as client_server_2,
    cloud as cloud,
    community as community,
    content_manifest as content_manifest,
    content_server as content_server,
    econ as econ,
    encrypted_app_ticket as encrypted_app_ticket,
    friend_messages as friend_messages,
    friends as friends,
    game_servers as game_servers,
    leaderboards as leaderboards,
    login as login,
    loyalty_rewards as loyalty_rewards,
    notifications as notifications,
    parental as parental,
    player as player,
    published_file as published_file,
    quest as quest,
    reviews as reviews,
    store as store,
    two_factor as two_factor,
    ucm as ucm,
    user_news as user_news,
    user_stats as user_stats,
)

[setattr(cls, "_betterproto", betterproto.ProtoClassMetadata(cls)) for cls in ProtobufMessage.__subclasses__()]
[setattr(cls, "_betterproto", betterproto.ProtoClassMetadata(cls)) for cls in UnifiedMessage.__subclasses__()]



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
from abc import ABCMeta
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, cast

from typing_extensions import Self, TypeVar

from ... import utils
from ...abc import BaseUser
from ...trade import Inventory, Item
from ...types.id import AssetID
from .enums import (
    ItemCustomizationNotification as ItemCustomizationNotificationEnum,
    ItemFlags,
    ItemOrigin,
    ItemQuality,
)
from .protobufs import base, econ, struct_messages

if TYPE_CHECKING:
    from datetime import datetime

    from ...abc import PartialUser
    from .client import ClientUser
    from .state import GCState

__all__ = (
    "Sticker",
    "Paint",
    "BaseItem",
    "CasketItem",
    "BaseInspectedItem",
    "InspectedItem",
    "BackpackItem",
    "Casket",
    "Backpack",
)


@dataclass(slots=True)
class Sticker:
    slot: Literal[0, 1, 2, 3, 4, 5] | None
    """The sticker's slot."""
    id: int
    """The sticker's ID."""
    wear: float | None = None
    """The sticker's wear."""
    rotation: float | None = None
    """The sticker's rotation."""
    scale: float | None = None
    """The sticker's scale."""
    tint_id: float | None = None
    """The sticker's tint_id."""

    offset_x: float = 0.0
    """The x offset of the sticker.

    .. versionadded:: 1.1.0
    """
    offset_y: float = 0.0
    """The y offset of the sticker.

    .. versionadded:: 1.1.0
    """


@dataclass(slots=True)
class Paint:
    """Represents the pain on an item."""

    index: float = 0.0
    """The paint's index."""
    seed: float = 0.0
    """The paint's seed."""
    wear: float = 0.0
    """The paint's wear."""
    offset_x: float = 0.0
    """The x offset of the paint.

    Warning
    -------
    Will always be ``0.0``

    .. versionadded:: 1.0.1
    """
    offset_y: float = 0.0
    """The y offset of the paint.

    Warning
    -------
    Will always be ``0.0``

    .. versionadded:: 1.0.1
    """

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} index={self.index} seed={self.seed} wear={self.wear}>"


class BaseItem(metaclass=ABCMeta):
    """Represents an item received from the Game Coordinator."""

    __slots__ = SLOTS = (
        "position",
        "paint",
        "tradable_after",
        "stickers",
        "kill_eater_value",
        "kill_eater_type",
        "_state",
        *tuple(base.Item.__annotations__),
    )
    if not TYPE_CHECKING:
        __slots__ = ()

    _state: GCState
    position: int
    """The item's position."""
    paint: Paint
    """The item's paint."""
    tradeable_after: datetime
    """The time the item's tradeable after."""
    kill_eater_value: int
    """The items kill eater value."""
    kill_eater_type: int
    """The type of kill eater attached."""
    quest_id: int
    """The quest this item was attained from."""
    stickers: list[Sticker]
    """The item's stickers."""
    id: AssetID
    """The item's asset ID."""
    account_id: int
    """The item's owner's 32-bit account ID."""
    inventory: int
    """Flags that aren't useful."""
    def_index: int
    """The item's def-index useful for its SKU."""
    quantity: int
    """The item's quantity."""
    level: int
    """The item's level."""
    quality: ItemQuality
    """The item's quality."""
    flags: ItemFlags
    """The item's flags."""
    origin: ItemOrigin
    """The item's origin."""
    custom_name: str
    """The item's custom name."""
    custom_description: str
    """The item's custom description."""
    attribute: list[base.ItemAttribute]
    """The item's attribute."""
    interior_item: base.Item
    """The item's interior item."""
    in_use: bool
    """Whether the item's in use."""
    style: int
    """The item's style."""
    original_id: int
    """The item's original ID."""
    equipped_state: list[base.ItemEquipped]
    """The item's equipped state."""
    rarity: int
    """The item's rarity."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} position={self.position}>"


@BaseItem.register
class CasketItem(BaseItem if TYPE_CHECKING else object):
    """Represents an item in a :class:`Casket`."""

    __slots__ = (*BaseItem.SLOTS, "_casket_id")
    _casket_id: AssetID

    @property
    def casket(self) -> Casket:
        """The casket this item is from."""
        backpack = self._state.backpack
        assert backpack is not None
        casket = utils.get(backpack, id=self._casket_id)
        assert isinstance(casket, Casket)
        return casket

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} casket={self.casket}>"


@dataclass
class _BaseInspectedItem(metaclass=ABCMeta):
    __slots__ = SLOTS = (
        "id",
        "def_index",
        "paint",
        "rarity",
        "quality",
        "kill_eater_score_type",
        "kill_eater_value",
        "custom_name",
        "stickers",
        "inventory",
        "origin",
        "quest_id",
        "drop_reason",
        "music_index",
        "ent_index",
    )
    if not TYPE_CHECKING:
        __slots__ = ()

    id: int
    """The item's asset ID."""
    def_index: int
    """The item's asset ID."""
    paint: Paint
    """The item's paint."""
    rarity: int
    """The item's rarity."""
    quality: ItemQuality
    """The item's quality."""
    kill_eater_score_type: int | None
    """The item's kill eater score type."""
    kill_eater_value: int | None
    """The item's kill eater value."""
    custom_name: str
    """The item's custom name."""
    stickers: list[Sticker]
    """The item's stickers."""
    inventory: int
    """The item's inventory."""
    origin: ItemOrigin
    """The item's origin."""
    quest_id: int
    """The item's quest id."""
    drop_reason: int
    """The item's drop reason."""
    music_index: int
    """The item's music index."""
    ent_index: int
    """The item's ent index."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id}>"


@dataclass(repr=False)
@_BaseInspectedItem.register
class BaseInspectedItem(_BaseInspectedItem if TYPE_CHECKING else object, metaclass=ABCMeta):
    """Represents an item received after inspecting an item."""

    __slots__ = _BaseInspectedItem.SLOTS
    __annotations__ = _BaseInspectedItem.__annotations__


OwnerT = TypeVar("OwnerT", bound="PartialUser", default="BaseUser", covariant=True)


@BaseInspectedItem.register
class InspectedItem(Item[OwnerT], _BaseInspectedItem):
    __slots__ = _BaseInspectedItem.SLOTS


F = TypeVar("F", bound=Callable[..., object])


def has_to_be_in_our_inventory(func: F) -> F:
    assert func.__doc__ is not None
    func.__doc__ += """

    Note
    ----
    For this method to work the item has to be in the client's backpack.
    """
    return func


class BackpackItem(Item[OwnerT], BaseItem):
    """A class to represent an item which can interact with the GC."""

    __slots__ = tuple(set(BaseItem.SLOTS) - {"_state"})
    _state: GCState

    REPR_ATTRS = (*Item.REPR_ATTRS, "position")

    @classmethod
    def from_item(cls, item: Item) -> Self:
        """A "type safe" way to cast ``item`` to a :class:`BackpackItem`."""
        return utils.update_class(item, cls.__new__(cls))

    async def rename_to(self: BackpackItem[ClientUser], name: str, tag: BackpackItem[ClientUser]) -> None:
        """Rename this item to ``name`` with ``tag``.

        Parameters
        ----------
        name
            The desired name.
        tag
            The tag to consume for this request.
        """
        future = self._state.ws.gc_wait_for(
            econ.ItemCustomizationNotification,
            check=lambda msg: (
                isinstance(msg, econ.ItemCustomizationNotification)
                and msg.request == ItemCustomizationNotificationEnum.NameItem
                and msg.item_id[0] == self.id
            ),
        )
        await self._state.ws.send_gc_message(
            struct_messages.NameItemRequest(name_tag_id=tag.id, item_id=self.id, name=name)
        )
        await future

    @has_to_be_in_our_inventory
    async def delete(self: BackpackItem[ClientUser]) -> None:
        """Delete this item."""
        await self._state.ws.send_gc_message(struct_messages.DeleteItemRequest(item_id=self.id))

    @property
    def inspect_url(self) -> str | None:
        """The inspect url of item if it's inspectable."""
        try:
            for action in self.actions:
                if "inspect" in action.name.lower():
                    return action.link.replace("%owner_steamid%", str(self.owner.id64)).replace(
                        "%assetid%", str(self.id)
                    )

        except (ValueError, KeyError):
            return None

    async def inspect(self) -> InspectedItem:
        """Inspect this item.

        Note
        ----
        This mutates ``self`` in a way that attributes available on the :class:`InspectedItem` are available on
        ``self``.
        """
        inspect_url = self.inspect_url
        if inspect_url is None:
            raise ValueError("Cannot inspect this item")
        basic = await self._state.client.inspect_item(url=inspect_url)
        return utils.update_class(self, basic)


class Casket(BackpackItem["ClientUser"]):
    """Represents a casket/storage container."""

    __slots__ = ("contained_item_count",)
    REPR_ATTRS = (*BackpackItem.REPR_ATTRS, "contained_item_count")

    contained_item_count: int
    """The number of items contained in the casket."""

    async def add(self, item: BackpackItem) -> None:
        """Add an item to this casket.

        Parameters
        ----------
        item
            The item to add.
        """
        future = self._state.ws.gc_wait_for(
            econ.ItemCustomizationNotification,
            check=lambda msg: (
                msg.request == ItemCustomizationNotificationEnum.CasketAdded and msg.item_id[0] == self.id
            ),
        )
        await self._state.ws.send_gc_message(econ.CasketItemAdd(casket_item_id=self.id, item_item_id=item.id))
        await future
        self.contained_item_count += 1

    async def remove(self, item: CasketItem) -> BackpackItem[ClientUser]:
        """Remove an item from this casket.

        Parameters
        ----------
        item
            The item to remove.

        Returns
        -------
        The item as a :class:`BackpackItem` in your inventory.
        """
        if item._casket_id != self.id:
            raise ValueError("item is not in this casket")

        future = self._state.ws.gc_wait_for(
            econ.ItemCustomizationNotification,
            check=lambda msg: (
                msg.request == ItemCustomizationNotificationEnum.CasketRemoved and msg.item_id[0] == self.id
            ),
        )
        await self._state.ws.send_gc_message(econ.CasketItemExtract(casket_item_id=self.id, item_item_id=item.id))
        await future
        self.contained_item_count -= 1

        return cast("BackpackItem[ClientUser]", await self._state.wait_for_item(item.id))

    async def contents(self) -> list[CasketItem]:
        """This casket's contents"""
        contained_items = [item for item in self._state.casket_items.values() if item._casket_id == self.id]
        if len(contained_items) == self.contained_item_count:
            return contained_items

        future = self._state.ws.gc_wait_for(
            econ.ItemCustomizationNotification,
            check=lambda msg: (
                msg.request == ItemCustomizationNotificationEnum.CasketContents and msg.item_id[0] == self.id
            ),
        )
        await self._state.ws.send_gc_message(econ.CasketItemLoadContents(casket_item_id=self.id, item_item_id=self.id))

        notification = await future
        return await asyncio.gather(
            *map(self._state.wait_for_casket_item, cast("list[AssetID]", notification.item_id[1:]))
        )

    async def rename_to(self, name: str) -> None:  # type: ignore
        """Rename this casket to ``name``.

        Parameters
        ----------
        name
            The name to rename the casket to.

        Note
        ----
        Caskets require names to work so if you've purchased one and forgot to activate it, use this method activate
        it.
        """
        # TODO consider this might need a lock to make sure that we can actually update the correct item
        await super().rename_to(name, _FakeNameTag())


class _FakeNameTag(BackpackItem["ClientUser"]):
    id = AssetID(0)
    __slots__ = ()

    def __init__(self, *_: object, **__: object):
        pass


class Backpack(Inventory[BackpackItem["ClientUser"], "ClientUser"]):
    """A class to represent the client's backpack."""

    @property
    def caskets(self) -> Sequence[Casket]:
        """The caskets in this backpack."""
        return [item for item in self if isinstance(item, Casket)]
